# -*- coding: utf-8 -*-

from nmpyc.model import model
from nmpyc.objective import objective
from nmpyc.opti import opti
from nmpyc.system import system
from nmpyc.constraints import constraints
from nmpyc.result import result
from nmpyc.nmpyc_array import *




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: Jonas Schiessl

import numpy as np

import matplotlib.pyplot as plt

from collections.abc import Iterable
import dill

import nmpyc as mpc

class result:
    """
    Class used to store the simulation results of the MPC simulation.

    To obtain the individual components of the simulation, 
    such as closed loop and open loop results, 
    the individual attributes need to be called.

    Also the result object contains information about errors and 
    other solver statistics, which can be used for further investigation 
    of the simulation progress.

    Additionally, this class provides a way to visualize the results 
    in a suitable way with the :py:meth:`~plot` method.
    
    Parameters
    ----------
    x0 : array
        Initial state.
    t0 : float
        Initial time.
    h : float
        Sampling rate.
    N : int
        MPC horizon.
    K : int
        Number of MPC Interations.

    """
    
    def __init__(self, x0, t0, h, N, K):
        
        self._t0 = t0
        self._x0 = x0
        
        self._t = t0
        
        self._h = h
        self._N = N
        self._K = K
        self._KK = K
        
        self._x_cl = []
        self._u_cl = []
        self._l_cl = []
        self._t_cl = []
        
        self._x_ol = []
        self._u_ol = []
        self._l_ol = []
        self._t_ol = []
        
        self._solver = None
        self._success = True
        self._error = None
        self._time = None
        self._time_ol = []
        
        self._default_kwargs = dict(xk = None,
                                    uk = None,
                                    iters = None,
                                    show_ol = False,
                                    phase1 = None,
                                    phase2 = None,
                                    dpi = 100,
                                    figsize = [8., 6.],
                                    usetex = True,
                                    fontsize = 14.,
                                    grid = True,
                                    linewidth = 2.,
                                    show_legend = True)
        
    @property 
    def t0(self):
        """float : Initial time"""
        return self._t0
    
    @property 
    def x0(self):
        """numpy.ndarray : Initial state."""
        return self._x0.A
    
    @property 
    def sampling_rate(self):
        """float : Sampling rate."""
        return self._h
    
    @property 
    def N(self):
        """int : MPC horizon"""
        return self._N
    
    @property
    def sucessfull_itertaions(self):
        """int : Number of succesfull MPC iterations."""
        return self._KK
    
    @property 
    def x_cl(self):
        """numpy.ndarray : Closed loop trajectory."""
        return self._x_cl.A
    
    @property 
    def u_cl(self):
        """numpy.ndarray : Closed loop feedback."""
        return self._u_cl.A
    
    @property 
    def l_cl(self):
        """numpy.ndarray : Stage costs evaluated at the 
        closed loop trajectory and feedback."""
        return self._l_cl.A
    
    @property 
    def t_cl(self):
        """numpy.ndarray : Time sequence at which the 
        closed loop states and controls are evaluated."""
        return self._t_cl.A
    
    @property 
    def x_ol(self):
        """list of numpy.ndarrays : List containing the 
        open loop trajectories of all MPC itertaions."""
        return self._x_ol
    
    @property 
    def u_ol(self):
        """list of numpy.ndarrays : List containing the 
        open loop optimal control values of all MPC itertaions."""
        return self._u_ol
        
    @property 
    def l_ol(self):
        """list of numpy.ndarrays : List containing the stage costs evaluated 
        at the open loop trajectories and controls of all MPC itertaions."""
        return self._l_ol
    
    @property
    def t_ol(self):
        """list of numpy.ndarrays : List containing the time sequences
        at which the closed loop states and controls are evaluated in 
        the open loop simulations."""
        return self._t_ol
    
    @property 
    def solver(self):
        """str : Name of the choosen optimization method."""
        return self._solver
    
    @property 
    def succes(self):
        """bool : True if the solver converged sucessfully in all MPC itertaions, 
        false if the MPC loop abort prematurely."""
        return self._succes
    
    @property 
    def error(self):
        """str : Error message with which the solver failed. 
        If success is True error is None."""
        return self._error
    
    @property 
    def ellapsed_time(self):
        """float : Total ellapsed time for the closed loop simulation."""
        return self._time
    
    @property 
    def ellapsed_time_per_itertaion(self):
        """list of float : List containing the ellapsed time of 
        every single itertaion of the closed loop simulation."""
        return self._time_ol
        
    def __str__(self):
        string = ''
        string += 'solver: ' + str(self._solver) + '\n'
        string += 'success: ' + str(self._success) + '\n'
        string += ('succesfull iterations: ' 
                   + str(self._K) 
                   + ' of ' 
                   + str(self._KK) 
                   + '\n')
        string += 'ellapsed time: ' + str(self._time) + '\n'
        string += 'x0: ' + str(self._x0.flatten()) + '\n'
        string += 't0: ' + str(self._t0) + '\n'
        string += 'h: ' + str(self._h) + '\n'
        string += 'N: ' + str(self._N) + '\n'
        string += 'x_cl: ' + str(self._x_cl) + '\n'
        string += 'u_cl: ' + str(self._u_cl) + '\n'
        string += 'l_cl: ' + str(self._l_cl.flatten())
        return string
        
    def _add_iteration(self, x, u, l):
        
        t = mpc.array(np.arange(self._t, 
                                self._t + (self._N + 1)*self._h - (self._h/2), 
                                self._h))
        self._t_ol += [t]
        
        self._t += self._h
        
        self._x_ol += [x]
        self._u_ol += [u]
        self._l_ol += [l]
        
    def _init_cl(self, t_end, x_end):
        
        for i in range(self._K):
            self._x_cl += [list(mpc.convert(self._x_ol[i][:,0], 
                                           'numpy').flatten())]
            self._u_cl += [list(mpc.convert(self._u_ol[i][:,0], 
                                           'numpy').flatten())]
            self._t_cl += [self._t_ol[i][0]]
            self._l_cl += [self._l_ol[i][0]]
        
        self._x_cl += [list(mpc.convert(x_end, 'numpy').flatten())]
        self._t_cl += [t_end]
        
        self._x_cl = mpc.array(self._x_cl).T
        self._u_cl = mpc.array(self._u_cl).T
        self._t_cl = mpc.array(self._t_cl)
        self._l_cl = mpc.array(self._l_cl)
        
        try:
            self._default_kwargs['xk'] = range(1, len(self._x_cl[:,0]) + 1)
            self._default_kwargs['uk'] = range(1, len(self._u_cl[:,0]) + 1)
            self._default_kwargs['iters'] = range(1, self._K + 1)
        except:
            self._default_kwargs['xk'] = []
            self._default_kwargs['uk'] = []
            
            
    def show_errors(self):
        """Shows the errors that occurred during the simualtion.
        
        For example, if the solver ipopt was selected and the 
        defined optimal control problem is infeasible, this method
        will print out the message 

        | *An error occured during itertaion 1 of 100:*
        | *Error in Opti::solve [OptiNode] at .../casadi/core/optistack.cpp:159:*
        | *.../casadi/core/optistack_internal.cpp:999: Assertion "return_success(accept_limit)" failed:*
        | *Solver failed. You may use opti.debug.value to investigate the latest values of variables. return_status is         'Infeasible_Problem_Detected'*

        For more informations about the error messages take a look 
        at the documentation of the respective solver.

        If the simulation was completly succesfull, the message 

        | *No error occured during the MPC-Loop*

        will be printed out.
        """
        
        if self._success:
            print('No error occured during the MPC-Loop.')
        else:
            print('An error occured during itertaion ' 
                  + str(self._K + 1) 
                  + ' of ' + str(self._KK) + ':')
            print(self._error)
        
    def plot(self, *args, **kwargs):
        """Plot the results of the MPC simulation.

        If no argument is passed, by default the closed loop states and controls are 
        plotted seperated in two subplots.

        If only a specific component of the solution should be plotted, 
        this can be customized by using a string as the first argument. 
        Valid arguments for this are

        - `state` for only plotting the closed loop state trajectories

        - `control` for only plotting the closed loop control values

        - `cost` for plotting the stage costs evaluated at the closed loop results

        - `phase` for plotting a phase portrait of two components of the solution.

        Further adjustments can be made with the help of the following keyword arguments.

        +----------------------+-------------------------------------------------+-------------------+
        | Argument             | Description                                     | Default value     |
        +======================+=================================================+===================+
        |xk                    | List specifying which components of the state   | `[1,...,nx]`      |
        |                      |                                                 |                   |
        |                      | are plotted.                                    |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |uk                    | List specifying which components of the control | `[1,...,nu]`      |
        |                      |                                                 |                   |
        |                      | are plotted.                                    |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |show_ol               |If True, the open loop simulation rersults are   | True              |
        |                      |                                                 |                   |
        |                      |also plotted additionaly to the closed loop      |                   |
        |                      |                                                 |                   |
        |                      |results.                                         |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |iters                 |List indicating from which iteration on the open | `[1,...,K+1]`     |
        |                      |                                                 |                   |
        |                      |loop results should be plotted.                  |                   |
        |                      |                                                 |                   |
        |                      |Will be ignored if show_ol is False.             |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |usetex                |If True, the captions are displayed in TEX style.| True              |
        +----------------------+-------------------------------------------------+-------------------+
        |grid                  |If True, a grid is displayed in the background   | True              |
        |                      |                                                 |                   |
        |                      |of the plot.                                     |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |show_legend           |If True, a legend will be displayed inside       | True              |
        |                      |                                                 |                   |
        |                      |the plot.                                        |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |phase1                |Phase 1 of the phase portrait plot.              | None              |
        |                      |                                                 |                   |
        |                      |Has the form `x_k` or `u_k`, where `k` determines|                   |
        |                      |                                                 |                   |
        |                      |the respective component.                        |                   |
        |                      |                                                 |                   |
        |                      |Will be ignored if args!='phase'.                |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |phase2                |Phase 2 of the phase portrait plot.              | None              |
        |                      |                                                 |                   |
        |                      |Has the form `x_k` or `u_k`, where `k` determines|                   |
        |                      |                                                 |                   |
        |                      |the respective component.                        |                   |
        |                      |                                                 |                   |
        |                      |Will be ignored if args!='phase'.                |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |dpi                   |resolution of the figure, see                    | 100               |
        +----------------------+-------------------------------------------------+-------------------+
        |figsize               |The size of the figure.                          | [8., 6.]          |
        |                      |                                                 |                   |
        |                      |Has the form  [width, height] in inches.         |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |linewidth             |Set the line width in points.                    | 2.                |
        +----------------------+-------------------------------------------------+-------------------+
        |fontsize              |The font size of the annotations.                | 14.               |
        |                      |                                                 |                   |
        |                      |If the value is numeric the size will be the     |                   |
        |                      |                                                 |                   |
        |                      |absolute font size in points.                    |                   |
        +----------------------+-------------------------------------------------+-------------------+

        """
        
        if self._default_kwargs['xk'] == []:
            print('Cant plot anything because there is' 
                  + ' no successfull itertaion stored.')
            return
        
        for item in kwargs.keys():
            if item not in self._default_kwargs.keys():
                raise KeyError()
                
        options = self._default_kwargs.copy()
        options.update(kwargs)
        
        if isinstance(options['xk'], int):
            options['xk'] = [options['xk']]
        if not isinstance(options['xk'], Iterable):
            raise TypeError('xk must be an integer or iterable')
            
        if isinstance(options['uk'], int):
            options['uk'] = [options['uk']]
        if not isinstance(options['uk'], Iterable):
            raise TypeError('uk must be an integer or iterable')
            
        if isinstance(options['iters'], int):
            options['iters'] = [options['iters']]
        if not isinstance(options['iters'], Iterable):
            raise TypeError('iters must be an integer or iterable')
        
        for item in options['xk']:
            if item not in self._default_kwargs['xk']:
                raise ValueError(
                    'Value ' + str(item) + ' of xk is out of range.')
                
        for item in options['uk']:
            if item not in self._default_kwargs['uk']:
                raise ValueError(
                    'Value ' + str(item) + ' of uk is out of range.')
                
        for item in options['iters']:
            if item not in self._default_kwargs['iters']:
                raise ValueError(
                    'Value ' + str(item) + ' of iters is out of range.')
        
        dimsx = [x-1 for x in options['xk']] 
        dimsu = [x-1 for x in options['uk']] 
        iters = [x-1 for x in options['iters']]
        
        plt.rcParams.update({
            "text.usetex": options['usetex'],
            "font.family": "sans-serif",
            "font.sans-serif": ["Helvetica"],
            "font.size": options['fontsize']})
        
        if len(args) == 0: 
            key = 'cl'
        elif len(args) == 1:
            if args[0] == 'state':
                key = 'x_'
            elif args[0] == 'control':
                key = 'u_'
            elif args[0] == 'phase':
                key = 'phase_'
            elif args[0] == 'cost':
                key = 'l_'    
            else: raise ValueError()
            
            if options['show_ol']:
                key += 'ol'
            else:
                key += 'cl'
        
        if key == 'cl':
            
            if options['usetex']:
                ylabelx = r'\textbf{state}'
                ylabelu = r'\textbf{control}'
                xlabel = r'\textbf{time}'
                title = r'\textbf{Closed Loop}'
            else:
                ylabelx = 'state'
                ylabelu = 'control'
                xlabel = 'time'
                title = 'Close Loop'
            
            fig, (ax1, ax2) = plt.subplots(2, 1, tight_layout=True, 
                                           sharex=True, 
                                           figsize=options['figsize'], 
                                           dpi=options['dpi'])
            fig.subplots_adjust(hspace = 0.2)
            
            for i in dimsx:
                if options['usetex']:
                    label = r'$x_' + str(i + 1) + '$'
                else:
                    label = 'x_' + str(i + 1)
                ax1.plot(self._t_cl, self._x_cl[i,:], 
                         label=label, linewidth=options['linewidth'])
            
            ax1.set_ylabel(ylabelx)
            ax1.grid(options['grid'])
            if options['show_legend']:
                ax1.legend()

            for i in dimsu:
                if options['usetex']:
                    label = r'$u_' + str(i + 1) + '$'
                else:
                    label = 'u_' + str(i + 1)
                ax2.step(self._t_cl[:-1], self._u_cl[i, :], where='post', 
                         label=label, linewidth=options['linewidth'])
            
            ax2.set_ylabel(ylabelu)
            ax2.grid(options['grid'])
            if options['show_legend']:
                ax2.legend()
            
            ax2.set_xlabel(xlabel)
            
            fig.suptitle(title)
            fig.align_ylabels((ax1, ax2))
            plt.show()

        elif key == 'x_cl':
            
            if options['usetex']:
                ylabelx = r'\textbf{state}'
                xlabel = r'\textbf{time}'
                title = r'\textbf{Closed Loop States}'
            else:
                ylabelx = 'state'
                xlabel = 'time'
                title = 'Close Loop States'
            
            fig, (ax1) = plt.subplots(1, 1, tight_layout=True, 
                                      figsize=options['figsize'], 
                                      dpi=options['dpi'])
            fig.subplots_adjust(hspace = 0.5)
            
            for i in dimsx:
                if options['usetex']:
                    label = r'$x_' + str(i + 1) + '$'
                else:
                    label = 'x_' + str(i + 1)
                ax1.plot(self._t_cl, self._x_cl[i, :], label=label, 
                         linewidth=options['linewidth'])
            
            ax1.set_xlabel(xlabel)
            ax1.set_ylabel(ylabelx)
            if options['show_legend']:
                ax1.legend()
            ax1.grid(options['grid'])
            
            fig.suptitle(title)
            plt.show()
            
        elif key == 'u_cl':
            
            if options['usetex']:
                ylabelu = r'\textbf{control}'
                xlabel = r'\textbf{time}'
                title = r'\textbf{Closed Loop Controls}'
            else:
                ylabelu = 'control'
                xlabel = 'time'
                title = 'Closed Loop Controls'
            
            fig, (ax1) = plt.subplots(1, 1, tight_layout = True, 
                                      figsize = options['figsize'], 
                                      dpi = options['dpi'])
            
            for i in dimsu:
                if options['usetex']:
                    label = r'$u_' + str(i + 1) + '$'
                else:
                    label = 'u_' + str(i + 1)
                ax1.step(self._t_cl[:-1], self._u_cl[i, :], where = 'post', 
                         label = label, linewidth = options['linewidth'])
            
            ax1.set_xlabel(xlabel)
            ax1.set_ylabel(ylabelu)
            if options['show_legend']:
                ax1.legend()
            ax1.grid(options['grid'])
            
            fig.suptitle(title)
            plt.show()
            
        elif key == 'x_ol':
            
            nx = len(self._x_cl[:,0])
            
            if options['usetex']:
                xlabel = r'\textbf{time}'
                title = r'\textbf{Open Loop States}'
            else:
                xlabel = 'time'
                title = 'Open Loop States'
            
            fig, ax = plt.subplots(len(dimsx), 1, sharex=True, 
                                   tight_layout = True, 
                                   figsize = options['figsize'], 
                                   dpi = options['dpi'])
            fig.subplots_adjust(hspace = 0.2)
            
            if nx == 1: 
                ax = (ax,)
            
            j = 0
            for k in dimsx:
                if options['usetex']:
                    label = r'$x_' + str(k + 1) + '$'
                    ylabelx = r'$\mathbf{x_' + str(k + 1) + '}$'
                else:
                    label = 'x_' + str(k + 1)
                    ylabelx = 'x_' + str(k + 1) 
                for i in iters:
                    ax[j].plot(self._t_ol[i], self._x_ol[i][k, :], 
                               label = label, color = 'grey', 
                               linewidth = options['linewidth'])
                
                ax[j].plot(self._t_cl, self._x_cl[k, :], 'o-', alpha = 0.3, 
                           label = label, color = 'b', 
                           linewidth = options['linewidth'])
                
                ax[j].set_ylabel(ylabelx)
                ax[j].grid(options['grid'])
                j += 1
                
            ax[-1].set_xlabel(xlabel)
            
            fig.suptitle(title)
            fig.align_ylabels(ax)
            plt.show()
            
        elif key == 'u_ol':
            
            nu = len(self._u_cl[:, 0])
            
            if options['usetex']:
                xlabel = r'\textbf{time}'
                title = r'\textbf{Open Loop States}'
            else:
                xlabel = 'time'
                title = 'Open Loop Controls'
            
            fig, ax = plt.subplots(len(dimsu), 1, sharex = True, 
                                   tight_layout = True, 
                                   figsize = options['figsize'], 
                                   dpi = options['dpi'])
            fig.subplots_adjust(hspace = 0.2)
            
            if nu == 1: 
                ax = (ax,)
            
            j = 0
            for k in dimsu:
                if options['usetex']:
                    label = r'$u_' + str(k + 1) + '$'
                    ylabelu = r'$\mathbf{u_' + str(k + 1) + '}$'
                else:
                    label = 'u_' + str(k + 1)
                    ylabelu = 'u_' + str(k + 1) 
                for i in iters:
                    ax[j].step(self._t_ol[i][:-1], self._u_ol[i][k, :], 
                               where = 'post', label = label, color = 'grey', 
                               linewidth = options['linewidth'])
                
                ax[j].step(self._t_cl[:-1], self._u_cl[k, :], 'o-', alpha = 0.3, 
                           where = 'post', label = label, color = 'b', 
                           linewidth = options['linewidth'])
                
                ax[j].set_ylabel(ylabelu)
                ax[j].grid(options['grid'])
                j += 1
                
            ax[-1].set_xlabel(xlabel)
            
            fig.suptitle(title)
            fig.align_ylabels(ax)
            plt.show()
            
        elif key == 'l_cl':
            
            if options['usetex']:
                xlabel = r'\textbf{time}'
                ylabel = r'$\mathbf{\ell (x,u)}$'
                title = r'\textbf{Closed Loop Costs}'
            else:
                xlabel = 'time'
                ylabel = 'l(x,u)'
                title = 'Closed Loop Costs'
                    
            fig, (ax1) = plt.subplots(1, 1, tight_layout = True, 
                                      figsize = options['figsize'], 
                                      dpi = options['dpi'])
            
            ax1.plot(self._t_cl[:-1], self._l_cl, 
                     linewidth = options['linewidth'])
            
            ax1.set_xlabel(xlabel)
            ax1.set_ylabel(ylabel)
            ax1.grid(options['grid'])
            
            fig.suptitle(title)
            plt.show()
                    
        elif key == 'l_ol':
            
            if options['usetex']:
                xlabel = r'\textbf{time}'
                ylabel = r'$\mathbf{\ell (x,u)}$'
                title = r'\textbf{Open Loop Costs}'
            else:
                xlabel = 'time'
                ylabel = 'l(x,u)'
                title = 'Open Loop Costs'
                    
            fig, (ax1) = plt.subplots(1, 1, tight_layout = True, 
                                      figsize = options['figsize'], 
                                      dpi = options['dpi'])
            
            for i in iters:
                ax1.plot(self._t_ol[i][:-1], self._l_ol[i], color='grey')
            
            ax1.plot(self._t_cl[:-1], self._l_cl, 'o-', alpha = 0.3, 
                     color = 'b', linewidth = options['linewidth'])
            
            ax1.set_xlabel(xlabel)
            ax1.set_ylabel(ylabel)
            ax1.grid(options['grid'])
            
            fig.suptitle(title)
            plt.show()
            
        elif key == 'phase_cl':
            
            phase = (kwargs['phase1'], kwargs['phase2'])
            
            if options['usetex']:
                xlabel = r'$\mathbf{' + phase[0] + '}$'
                ylabel = r'$\mathbf{' + phase[1] + '}$'
                title = (r'$\mathbf{' + phase[0] + '-' + phase[1] 
                         + '~Closed~Loop' + '~Portrait' + '}$')
            else:
                xlabel = phase[0]
                ylabel = phase[1]
                title = (phase[0] + '-' + phase[1] 
                         + ' Closed Loop' + ' Portrait')
            
            if phase[0][0] == 'u' or phase[1][0] == 'u':
                T = self._K
            else:
                T = self._K + 1
            
            x = []
            
            for i in range(2):
                if phase[i][1] != '_': raise ValueError()
                if phase[i][0] == 'x':
                    k = int(phase[i][2:])
                    x += [self._x_cl[k-1, :T]]
                if phase[i][0] == 'u':
                    k = int(phase[i][2:])
                    x += [self._u_cl[k-1, :]]
                    
            fig, (ax1) = plt.subplots(1, 1, tight_layout = True, 
                                      figsize = options['figsize'], 
                                      dpi = options['dpi'])
            
            ax1.plot(x[0], x[1], linewidth = options['linewidth'])
            
            ax1.set_xlabel(xlabel)
            ax1.set_ylabel(ylabel)
            ax1.grid(options['grid'])
            
            fig.suptitle(title)
            plt.show()
                    
        elif key == 'phase_ol':
             
            phase = (kwargs['phase1'], kwargs['phase2'])
            
            if options['usetex']:
                xlabel = r'$\mathbf{' + phase[0] + '}$'
                ylabel = r'$\mathbf{' + phase[1] + '}$'
                title = (r'$\mathbf{' + phase[0] + '-' + phase[1] 
                         + '~Open~Loop' + '~Portrait' + '}$')
            else:
                xlabel = phase[0]
                ylabel = phase[1]
                title = phase[0] + '-' + phase[1] + ' Open Loop' + ' Portrait'
            
            x = []
            
            if phase[0][0] == 'u' or phase[1][0] == 'u':
                T1 = self._N
                T2 = self._K
            else:
                T1 = self._N + 1
                T2 = self._K + 1
            
            fig, (ax1) = plt.subplots(1, 1, tight_layout = True, 
                                      figsize = options['figsize'], 
                                      dpi = options['dpi'])
            
            for j in iters:
                for i in range(2):
                    if phase[i][1] != '_': 
                        raise ValueError()
                    if phase[i][0] == 'x':
                        k = int(phase[i][2:])
                        x += [self._x_ol[j][k - 1, :T1]]
                    if phase[i][0] == 'u':
                        k = int(phase[i][2:])
                        x += [self.u_ol[j][k - 1, :]]
                
                ax1.plot(x[0], x[1], color = 'grey', 
                         linewidth = options['linewidth'])
                x = []
                        
            for i in range(2):
                if phase[i][0] == 'x':
                    k = int(phase[i][2:])
                    x += [self._x_cl[k - 1, :T2]]
                if phase[i][0] == 'u':
                    k = int(phase[i][2:])
                    x += [self._u_cl[k - 1, :]]
            
            ax1.plot(x[0], x[1], 'o-', alpha = 0.3, color = 'b', 
                     linewidth = options['linewidth'])
            
            ax1.set_xlabel(xlabel)
            ax1.set_ylabel(ylabel)
            ax1.grid(options['grid'])
            
            fig.suptitle(title)
            plt.show()
    
    def save(self, path):
        """Saving the result to a given file with `dill <https://dill.readthedocs.io/en/latest/dill.html>`_.
        
        The path can be absolut or relative and 
        the ending of the file is arbitrary.

        Parameters
        ----------
        path : str
            String defining the path to the desired file. 



        For example

        >>> result.save('result.pickle')
        
        will create a file `result.pickle` containing the nMPyC result object.
        """
        
        with open(path, "wb") as output_file:
            dill.dump(self, output_file, -1)
    
    @classmethod
    def load(cls, path):
        """Loads a nMPyC result object from a file.
        
        The specified path must lead to a file that was previously saved with
        :py:meth:`~save`.
        
        Parameters
        ----------
        path : str
            String defining the path to the file containing the nMPyC 
            result object. 
            
            
        For example

        >>> result.load('result.pickle')
        
        will load the result previously saved with :py:meth:`~save`.
            
        """
        
        try:
            with open(path, "rb") as input_file:
                e = dill.load(input_file)
        except:
            raise Exception(
                'Can not load result from file. File not readable!')
            
        if not isinstance(e, result):
            raise Exception(
                'Can not load result from file. File does not cotain a result!')
            
        return e
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: Jonas Schiessl

import nmpyc as mpc
from nmpyc.opti import opti

import time
import dill

class model:
    """
    Class that contains all the components of the optimal control problem.
    
    Can be used to perform open and closed loop simulations.
    
    Parameters
    ----------
    objective : objective
        nMPyC-objective defining the objective of the 
        optimal control problem.
    system : system
        nMPyC-system defining the system dynamics of the 
        optimal control problem.
    constraints : constraints optional
        nMPyC-constraints defining the constraints of the optimal control 
        problem. If constraints is None the problem is unconstrained. 
        The default is None.

    """
    
    def __init__(self, objective, system, constraints=None):
        
        
        self.objective = objective
        self.system = system   
        self.constraints = constraints
        self._opti = opti()
        self._N = None
        
    @property 
    def objective(self):
        """objective : Objective of the optimal control problem."""
        return self._objective
        
    @objective.setter
    def objective(self,obj):
        if isinstance(obj, mpc.objective):
            self._objective = obj
        else:
            raise TypeError(
                'objective must be of type objective - not ' 
                + str(type(obj)))
            
    @property
    def system(self):
        """system : System dynamics of the optimal control problem."""
        return self._system
    
    @system.setter
    def system(self,syst):
        if isinstance(syst, mpc.system):
            self._system = syst
            self._nx = self._system.nx
            self._nu = self._system.nu
        else:
            raise TypeError(
                'system must be of type system - not ' 
                + str(type(syst)))
            
    @property
    def constraints(self):
        """constraints : Constraints of the optimal control problem."""
        return self._constraints
    
    @constraints.setter
    def constraints(self,cons):
        if cons is None:
            cons = mpc.constraints()
        if isinstance(cons, mpc.constraints):
            self._constraints = cons
        else:
            raise TypeError(
                'constraints must be of type constraints - not ' 
                + str(type(cons)))
    
    @property
    def opti(self):
        """opti : Optimizer for the optimal control problem.
        
        This property can be used to set different optimization 
        options.
        A distinction is made between basic settings of the 
        optimizer and solver-specific settings.

        The basic settings can be adjusted by calling 

        >>> model.opti.set_options({..})

        The dictionary that is passed can contain the following entries

        +----------------------+-------------------------------------------------+-------------------+
        |Parameter             | Description                                     | Default value     |
        +======================+=================================================+===================+
        |solver                | String defining which solver is for             | auto              |
        |                      |                                                 |                   |
        |                      | optimization. Currently supported solvers are   |                   |
        |                      |                                                 |                   |
        |                      | - ipotpt                                        |                   |
        |                      |                                                 |                   |
        |                      | - sqpmethod                                     |                   |
        |                      |                                                 |                   |
        |                      | - ospq                                          |                   |
        |                      |                                                 |                   |
        |                      | - SLSQP                                         |                   |
        |                      |                                                 |                   |
        |                      | - trust-constr                                  |                   |
        |                      |                                                 |                   |
        |                      | For auto a suitable solver depending on other   |                   |
        |                      |                                                 |                   |
        |                      | options and parameters is selected.             |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |full_discretization   |If True, the method of full discretiziation is   | True              |
        |                      |                                                 |                   |
        |                      |used for optimization. Otherwise the the system  |                   |
        |                      |                                                 |                   |
        |                      |dynamics is resolved in the objective function.  |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |tol                   |The toleranz of the solver.                      | 1e-06             |
        |                      |                                                 |                   |
        |                      |If the solver distinguishes between relative     |                   |
        |                      |                                                 |                   |
        |                      |and absolute tolerances, both are set to this    |                   |
        |                      |                                                 |                   |
        |                      |value.                                           |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |maxiter               |Maximal number of iterations during the          | 5000              |
        |                      |                                                 |                   |
        |                      |optimization progress.                           |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |verbose               |If True, the verbose option of the selected      | False             |
        |                      |                                                 |                   |
        |                      |solver ist activated.                            |                   |
        +----------------------+-------------------------------------------------+-------------------+
        |initial_guess         |Initial guess for the optimization variable u.   | None              |
        |                      |                                                 |                   |
        |                      |Must be an array of shape (nx,N).                |                   |
        |                      |                                                 |                   |
        |                      |If the initial guess has not the right shape or  |                   |
        |                      |                                                 |                   |
        |                      |is None it will be set to nmpyc.ones((nu,N))*0.1 |                   |
        |                      |                                                 |                   |
        |                      |by default.                                      |                   |
        +----------------------+-------------------------------------------------+-------------------+

        The auto option of the solver selection follows the rule

        1. If the optimal control problem is recognized as a LQP and a fixed step discretization of the system is given, osqp is selected.

        2. If a condition of 1. is violated and not a SciPy discretization method is choosen, ipopt is selected.

        3. Otherwise SLSQP is selected.

        The solver-specific settings can be custamized by calling

        >>> model.opt.set_solverOptions({..})

        Valid parameters which the passed dictionary can contain 
        are depending on the selected solver.
        For a list of these settings take a look at 

        - `Sourceforge <http://casadi.sourceforge.net/v2.0.0/api/html/d6/d07/classcasadi_1_1NlpSolver.html>`_ for the CasADi solvers

        - `SciPy Documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_ for the SciPy solvers

        - `OSQP Website <https://osqp.org/docs/interfaces/solver_settings.html>`_ for the osqp solver
        
        """
        return self._opti
    
    @property
    def N(self):
        """int : Prediction horizon of the MPC loop."""
        return self._N
    
    @N.setter
    def N(self,NN):
        if isinstance(NN, int):
            if NN <= 0 : 
                raise ValueError('MPC horizon N must be greater than zero')
            self._N = NN
        else: 
            raise TypeError(
                'MPC horizon N must be of type integer - not ' 
                + str(type(NN)))
        
    def __str__(self):
        
        string = 'OBJECTIVE: \n'
        string += str(self.objective) + '\n'
        string += 'SYSTEM: \n'
        string += str(self.system) + '\n'
        string += 'CONSTRAINTS: \n'
        string += str(self.constraints)
        
        return string
    
    def solve_ocp(self, x0, N, discount=None):
        """Solves the finit horizon optimal control problem.

        Parameters
        ----------
        x0 : array
            Initial value of the optimal control problem.
        N : int
            Prediction horizon of the control problem.
        discount : float, optional
            Discountfactor of the objective. The default is None.

        Returns
        -------
        u_ol : array
            Optimal control sequence.
        x_ol : array
            Optimal trajectory corresponding to the optimal control sequence.

        """
        self.N = N
        
        self._objective.discont = discount
        
        if isinstance(x0, mpc.array):
            if x0.dim != (self._nx,1):
                raise ValueError(
                    'x0 has the wrong dimension - ' 
                    + str(x0.dim) 
                    + ' != (' + str(self._nx) 
                    + ',1)')
        else: 
            raise TypeError(
                'x0 must be of type array - not ' 
                + str(type(x0)))
        
        self.opti.init_solver(self.objective, self.system, 
                              self.constraints, self.N)
        
        u_ol, x_ol = self.opti.solve(self.system.t0,x0)
        
        return (u_ol, x_ol)
            
    def mpc(self, x0, N, K, discount=None):
        """Solves the optimal control problem via model predictive control.

        Parameters
        ----------
        x0 : array
            Initial state of the optimal control problem.
        N : int
            MPC horizon.
        K : int
            Number of MPC itertaions.
        discount : float, optional
            Discountfactor of the objective. The default is None.

        Returns
        -------
        res : result
            nMPyC result object containing the optimiaztion results of 
            the closed and open loop simulations.

        """
        
        self.N = N
        self._objective.discont = discount
        
        if isinstance(K, int):
            if K <= 0 : 
                raise ValueError(
                    'Number of Iterations K must be greater than zero')
        else: 
            raise TypeError(
                'Number of Iterations K must be of type integer - not ' 
                + str(type(K)))
        
        if isinstance(x0, mpc.array):
            if x0.dim != (self._nx, 1):
                raise ValueError(
                    'x0 has the wrong dimension - ' 
                    + str(x0.dim) + ' != (' + str(self._nx) + ',1)')
            x = x0
        else: 
            raise TypeError(
                'x0 must be of type array - not ' 
                + str(type(x0)))
        
        k = 0   # set counter to zero
        
        res = mpc.result(x0, self.system.t0, self.system.h, N, K)
        
        print('Initialize solver ...')
        self.opti.init_solver(self.objective, self.system, 
                              self.constraints, self.N)
        print('... DONE')
        
        print("STARTING MPC-LOOP: (initial value: " + str(x0.flatten()) + ")", 
              flush = True)

        # --- MPC loop ---

        t_s_total = time.time() #timepoint for total time messurement
        
        t = self.system.t0

        while k < K:
            # solve MPC optimal control problem
            t_s_step = time.time() 
            
            # solve ocp
            u_ol, x_ol = self.opti.solve(t, x)
            
            #proof if optimizer terminated sucessfully
            if u_ol is None:
                res._success = False
                res._error = x_ol
                res._K = k
                print('ITERATION ENDED UNSUCCESSFULLY!')
                break

            # get control to the right shape for getting MPC feedback
            u = u_ol[:,0]

            # apply MPC feedback to the system
            x = self.system.system_discrete(t, x, u)
            
            #calculate costs
            l_ol = []
            t_l = t
            for l in range(self.N):
                t_l += self.system.h
                l_ol += [float(self.objective.stagecosts(t_l, x_ol[:,l], u_ol[:,l]))]
            l_ol = mpc.array(l_ol)
            
            #save current itertaion
            res._add_iteration(x_ol, u_ol, l_ol)
            
            # calculate time ellapsed in the iteration step
            elapsed_step = time.time() - t_s_step
            res._time_ol += [elapsed_step]
            
            print("... finished iteration " 
                  + str(k+1) 
                  +" of "
                  + str(K)
                  + " (time: " + str(elapsed_step) + ")", flush = True)
            print("   ... feedback: " + str(u.flatten()), flush = True)
            print("   ... closed loop state: " + str(x.flatten()), 
                  flush = True)
            
            k += 1 # next itertion
            t += self.system.h # next timestep
            
        # calculate total time ellapsed
        elapsed_total = time.time() - t_s_total
        res._time = elapsed_total
        
        res._solver = self._opti.solver
        
        print("END" + " (total time: " + str(elapsed_total) + ")", 
              flush = True)
        
        res._init_cl(t, x)
        
        return res
    
    def save(self, path):
        """Saving the model to a given file with `dill <https://dill.readthedocs.io/en/latest/dill.html>`_.
        
        The path can be absolut or relative and 
        the ending of the file is arbitrary.

        Parameters
        ----------
        path : str
            String defining the path to the desired file. 



        For example

        >>> model.save('objective.pickle')
        
        will create a file `model.pickle` containing the nMPyC model object.
        """
        
        self._opti._delete_optistack()
        with open(path, "wb") as output_file:
            dill.dump(self, output_file, -1)
    
    @classmethod
    def load(cls, path):
        """Loads a nMPyC model object from a file.
        
        The specified path must lead to a file that was previously saved with
        :py:meth:`~save`.
        
        Parameters
        ----------
        path : str
            String defining the path to the file containing the nMPyC 
            model object. 
            
            
        For example

        >>> model.load('model.pickle')
        
        will load the model previously saved with :py:meth:`~save`.
            
        """
        
        try:
            with open(path, "rb") as input_file:
                e = dill.load(input_file)
        except:
            raise Exception(
                'Can not load model from file. File not readable!')
            
        if not isinstance(e, model):
            raise Exception(
                'Can not load model from file. File does not cotain a model!')
            
        return e
            
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: Jonas Schiessl

import casadi as cas
import numpy as np

import osqp

import nmpyc as mpc
from nmpyc.utils import flat_list

from scipy.optimize import minimize
import scipy.sparse as sparse
from scipy.linalg import block_diag
from numpy.linalg import matrix_power

class opti:
    """
    A class used to sore the optimization problem of the OCP.
    
    In this class the choosen optimizer is initialized and the minimization 
    method is called if the problem should bes solved durcing the MPC loop.
    """
    
    def __init__(self):
        
        self._solver = 'auto'
        self._full_discretization = True 
        self._tol = 1e-06
        self._maxiter = 5000
        self._verbose = False
        
        self._U_start = None
        
        self._options={}
    
    @property
    def solver(self):
        """str : Used solver for optimization."""
        return self._method
    
    @solver.setter 
    def solver(self,value):
        if value == 'osqp':
            self._solver = 'osqp'
        elif value in ['sqpmethod','ipopt']:
            self._solver = 'casadi'
        elif value in ['SLSQP', 'trust-constr']:
            self._solver = 'scipy'
        elif value == 'auto':
            self._solver = value
        else: 
            raise ValueError(value + ' is not a availible solver')
            
        self._method = value
        
    @property 
    def full_discretization(self):
        """bool : If true full discretization is applied to the problem."""
        return self._full_discretization
    
    @full_discretization.setter 
    def full_discretization(self, value):
        if isinstance(value, bool):
            self._full_discretization = value
        else:
            raise TypeError(
                'full_discretization must be of type boolean - not ' 
                + str(type(value)))
            
        self._full_discretization = value
            
    @property 
    def tol(self):
        """float : Tolerance of the optimizer."""
        return self._tol
    
    @tol.setter 
    def tol(self, value):
        if not isinstance(value, (int,float)):
            raise TypeError(
                'tol must be of type int or float - not ' 
                + str(type(value)))
        
        if value <= 0:
            raise ValueError(
                'tol must be grater than zero - you have ' 
                + str(value))
            
        self._tol = value
        
    @property 
    def maxiter(self):
        """int : Maximal number of itertaions during the optimization."""
        return self._maxiter
    
    @maxiter.setter
    def maxiter(self, value):
        if isinstance(value, float):
            if not value.is_integer():
                raise TypeError(
                    'maxiter must be of type integer - not ' 
                    + str(type(value)))
            else:
                value = int(value)
                
        if not isinstance(value, int):
            raise TypeError(
                    'maxiter must be of type integer - not ' 
                    + str(type(value)))
            
        if value <= 0:
            raise ValueError(
                'maxiter must be grater than zero - you have ' 
                + str(value))
            
        self._maxiter = value
        
    @property 
    def verbose(self):
        """bool : If true the verbose option of the optimizer is activated."""
        return self._verbose
    
    @verbose.setter 
    def verbose(self, value):
        if isinstance(value, bool):
            self._full_discretization = value
        else:
            raise TypeError(
                'verbose must be of type boolean - not ' 
                + str(type(value)))
            
        self._verbose = value
            
    @property 
    def U_start(self, value):
        """array : Initial guess for the optimizer."""

        if not isinstance(value, mpc.array):
            raise TypeError(
                'inital guess must be of type array - not ' 
                + str(type(value)))
        
        if value.symbolic:
            raise TypeError(
                'inital guess mhas to be purely numeric,' 
                + ' but also has symbolic entries.')

        return self._U_start
    
    #TO-DO: proof initial guess
    @U_start.setter 
    def U_start(self,value):
        self._U_start = value
        
    @property 
    def solver_options(self):
        """dict : Dictionary containing the solver dependent options"""
        return self._options
    
    def _delete_optistack(self):
        self._optistack = None
        self._X = None
        self._U = None
        self._p = None
        self._t = None  
        
    def set_options(self, opts):
        """Set the solver independent options.

        Parameters
        ----------
        opts : dict
            Dictionary containing the keywords of the required options 
            and their values.
            
        """
        
        if not isinstance(opts, dict):
            raise TypeError(
                'opts must be of type dictionary - not ' + str(type(opts)))
        
        for key in opts.keys():
            if key in ['solver', 'full_discretization','tol', 
                       'maxiter', 'verbose', 'initial_guess']:
                if key == 'solver': 
                    self.solver = opts['solver']
                elif key == 'full_discretization': 
                    self.full_discretization = opts['full_discretization']
                elif key == 'tol':
                    self.tol = opts['tol']
                elif key == 'maxiter':
                    self.maxiter = opts['maxiter']
                elif key == 'verbose':
                    self.verbose = opts['verbose']
                elif key == 'initial_guess':
                    self.U_start = opts['initial_guess']
            else: 
                raise KeyError(key + ' is not a valid option')
            
    def set_solverOptions(self, options):
        """Set the solver dependent options.

        Parameters
        ----------
        opts : dict
            Dictionary containing the keywords of the required options 
            and their values.
            
        """
        
        if not isinstance(options, dict):
            raise TypeError(
                'opts must be of type dictionary - not ' + str(type(options)))
            
        self._options = options             
        
    def init_solver(self, objective, system, constraints, N):
        """Initalize the solver for the OCP defined by the input Parameters.

        Parameters
        ----------
        objective : objective
            Objective of the optimal control problem.
        system : system
            System of the optimal control problem.
        constraints : constraint
            Constraints of the optiomal control problem.
        N : int
            Prediction horizon for the finit horizon optimal control problem.

        """
        
        if isinstance(objective, mpc.objective):
            self._objective = objective
        else:
            raise TypeError(
                'objective must be of type objective - not ' 
                + str(type(objective)))
        
        if isinstance(system, mpc.system):
            self._system = system
        else:
            raise TypeError(
                'system must be of type system - not ' 
                + str(type(system)))
         
        if isinstance(constraints, mpc.constraints):
            self._constraints = constraints
        else:
            raise TypeError(
                'constraints must be of type constraints - not ' 
                + str(type(constraints)))
        
        if isinstance(N, int):
            if N <= 0 : 
                raise ValueError('MPC horizon N must be greater than zero')
            self._N = N
        else: 
            raise TypeError(
                'MPC horizon N must be of type integer - not ' + str(type(N)))
        
        self._nx = system.nx
        self._nu = system.nu
        
        self._t0 = self._system.t0
        
        if self._constraints.lower_bndx is None:
            self._constraints.add_bound('lower', 'state', 
                                        mpc.ones(self._nx) * (-mpc.inf))
        if self._constraints.upper_bndx is None:
            self._constraints.add_bound('upper', 'state', 
                                        mpc.ones(self._nx) * (mpc.inf))
        if self._constraints.lower_bndu is None:
            self._constraints.add_bound('lower', 'control', 
                                        mpc.ones(self._nu) * (-mpc.inf))
        if self._constraints.upper_bndu is None:
            self._constraints.add_bound('upper', 'control', 
                                        mpc.ones(self._nu) * (mpc.inf))
        if self._constraints.lower_bndend is None:
            self._constraints.add_bound('lower', 'terminal', 
                                        self._constraints.lower_bndx)
        if self._constraints.upper_bndend is None:
            self._constraints.add_bound('upper', 'terminal', 
                                        self._constraints.upper_bndx)
        
        if self._constraints.lower_bndx.dim[0] != self._nx:
            raise ValueError(
                'lower bound for state has the wrong dimension - ' 
                + str(self._constraints.lower_bndx.dim[0]) 
                + '!=' 
                + str(self._nx))
        if self._constraints.upper_bndx.dim[0] != self._nx:
            raise ValueError(
                'upper bound for state has the wrong dimension - ' 
                + str(self._constraints.upper_bndx.dim[0]) 
                + '!=' 
                + str(self._nx))
        if self._constraints.lower_bndu.dim[0] != self._nu:
            raise ValueError(
                'lower bound for control has the wrong dimension - ' 
                + str(self._constraints.lower_bndu.dim[0]) 
                + '!=' 
                + str(self._nu))
        if self._constraints.upper_bndu.dim[0] != self._nu: 
            raise ValueError(
                'upper bound for control has the wrong dimension - ' 
                + str(self._constraints.upper_bndu.dim[0]) 
                + '!=' 
                + str(self._nu))
        if self._constraints.lower_bndend.dim[0] != self._nx:
            raise ValueError(
                'lower bound for terminal state has the wrong dimension - ' 
                + str(self.constraints.lower_bndend.dim[0]) 
                + '!=' 
                + str(self._nx))
        if self._constraints.upper_bndend.dim[0] != self._nx:
            raise ValueError(
                'upper bound for terminal state has the wrong dimension - ' 
                + str(self.constraints.upper_bndend.dim[0]) 
                + '!=' 
                + str(self._nx))
        
        # Set initial guess for the control by default to 0.1
        if self._U_start is None:
            self._U_start = mpc.ones((self._nu, self._N))*0.1
        if (self._U_start.dim[0] != self._nu or 
            self._U_start.dim[1] != self._N):
            self._U_start = mpc.ones((self._nu, self._N))*0.1
            print('Warning: Initial guess has the wrong dimension.' 
                  + ' Changing to default vaulue.')
        self._X_start = mpc.array((self._nx, self._N + 1))
        
        if self._solver == 'auto':
            if (self._system.type == 'LQP' and 
                self._objective.type == 'LQP' and 
                self._constraints.type == 'LQP'):
                self._solver = 'osqp'
                self._method = 'osqp'
            elif self._system._integrator != 'scipy':
                self._solver = 'casadi'
                self._method = 'ipopt'
            else:
                self._solver = 'scipy'
                self._method = 'SLSQP'
            
        if self._solver == 'osqp':
            if (self._system.type == 'NLP' or 
                self._objective.type == 'NLP' or 
                self._constraints.type == 'NLP'):
                raise ValueError('osqp could only be applied to LQPs')
            self._init_osqp()
            
        elif self._solver == 'casadi':
            if self._system._integrator == 'scipy':
                raise ValueError(
                    'casadi could not be used with a scipy-integrator')
            self._init_casadi()
            
        elif self._solver == 'scipy':
            self._init_scipy()
        
        else: 
            raise ValueError(self._solver + 'is not a valid solver')
            
    def _init_osqp(self):
        
        self._A = None
        self._u = None
        self._l = None
        
        Q = mpc.convert(self._objective.stagecost[0],'numpy')
        R = mpc.convert(self._objective.stagecost[1],'numpy')
        NN = mpc.convert(self._objective.stagecost[2],'numpy')
        PP = mpc.convert(self._objective.terminalcost,'numpy')
        
        if self._full_discretization:
            self._A = np.eye(self._nx*(self._N+1)+self._nu*self._N)
            GG = np.block([np.eye(self._nx),
                           np.zeros((self._nx, (self._nu+self._nx)*self._N))])
        
            for i in range(self._N):
                G = np.block([np.zeros((self._nx,(i)*(self._nx+self._nu))),
                              mpc.convert(self._system.f[0]),
                              mpc.convert(self._system.f[1]), 
                              -np.eye(self._nx), 
                              np.zeros((self._nx,(self._N - i - 1)
                                        *(self._nx+self._nu)))])
                GG = np.vstack((GG, G))
            
            A = np.vstack((GG, self._A))
            
            self._l = np.zeros(self._nx*(self._N+1)*2 + self._nu*self._N)
            self._u = np.zeros(self._nx*(self._N+1)*2 + self._nu*self._N)
            for i in range(self._N):
                li = self._nx*(self._N + 1) + (i + 1)*self._nx + i*self._nu
                ui = (self._nx*(self._N + 1) 
                      + (i + 1)*self._nx 
                      + (i + 1)*self._nu)
                self._l[li:ui] = mpc.convert(self._constraints.lower_bndu, 
                                             'numpy').flatten()
                li = self._nx*(self._N + 1) + i*self._nx + i*self._nu
                ui = self._nx*(self._N + 1) + (i + 1)*self._nx + i*self._nu
                self._l[li:ui] = mpc.convert(self._constraints.lower_bndx, 
                                             'numpy').flatten()
            for i in range(self._N):
                li = self._nx*(self._N + 1) + (i + 1)*self._nx + i*self._nu
                ui = (self._nx*(self._N + 1) 
                      + (i + 1)*self._nx 
                      + (i + 1)*self._nu)
                self._u[li:ui] = mpc.convert(self._constraints.upper_bndu, 
                                             'numpy').flatten()
                li = self._nx*(self._N + 1) + i*self._nx + i*self._nu
                ui = self._nx*(self._N + 1) + (i + 1)*self._nx + i*self._nu
                self._u[li:ui] = mpc.convert(self._constraints.upper_bndx, 
                                             'numpy').flatten()
            self._l[-self._nx:] = mpc.convert(self._constraints.lower_bndend, 
                                              'numpy').flatten()
            self._u[-self._nx:] = mpc.convert(self._constraints.upper_bndend, 
                                              'numpy').flatten()
            
            for cons in self._constraints.linear_constr['eq']:
                GG = np.block([mpc.convert(cons[0]),
                               mpc.convert(cons[1]),
                               np.zeros((len(cons[0][:,0]),
                                         ((self._nu + self._nx)*(self._N - 1) 
                                          + self._nx)))])
                self._l = np.hstack((self._l,
                                     mpc.convert(cons[2]).flatten()))
                self._u = np.hstack((self._u,
                                     mpc.convert(cons[2]).flatten()))
                for i in range(self._N - 1):
                    G = np.block([np.zeros((len(cons[0][:,0]),
                                            (i + 1)*(self._nx+self._nu))),
                                  mpc.convert(cons[0]),
                                  mpc.convert(cons[1]),
                                  np.zeros((len(cons[0][:,0]),
                                            (self._nx
                                             +((self._N - i - 2)
                                               *(self._nx + self._nu)))))])
                    GG = np.vstack((GG, G))
                    self._l = np.hstack((self._l, 
                                         mpc.convert(cons[2]).flatten()))
                    self._u = np.hstack((self._u, 
                                         mpc.convert(cons[2]).flatten()))
                    
                A = np.vstack((A, GG))
                
            
            for cons in self._constraints.linear_constr['ineq']:
                GG = np.block([mpc.convert(cons[0]),
                               mpc.convert(cons[1]),
                               np.zeros((len(cons[0][:,0]),
                                         ((self._nu + self._nx)*(self._N - 1)
                                          + self._nx)))])
                self._l = np.hstack((self._l, mpc.convert(cons[2]).flatten()))
                for i in range(self._N - 1):
                    G = np.block([np.zeros((len(cons[0][:,0]),
                                            (i + 1)*(self._nx + self._nu))),
                                  mpc.convert(cons[0]),
                                  mpc.convert(cons[1]),
                                  np.zeros((len(cons[0][:,0]),
                                            (self._nx
                                             + ((self._N - i - 2)
                                               *(self._nx + self._nu)))))])
                    GG = np.vstack((GG, G))
                    self._l = np.hstack((self._l,
                                         mpc.convert(cons[2]).flatten()))
                
                A = np.vstack((A, GG))
                self._u = np.hstack((self._u, np.ones(len(GG[:,0]))*np.inf))
            
            for cons in self._constraints.linear_constr['terminal_eq']:
                GG = np.block([np.zeros((len(cons[0][:,0]),
                                         (self._nu + self._nx)*(self._N))),
                               mpc.convert(cons[0])])
                
                A = np.vstack((A, GG))
                self._l = np.hstack((self._l,
                                     mpc.convert(cons[1]).flatten()))
                self._u = np.hstack((self._u, 
                                     mpc.convert(cons[1]).flatten()))
                
            for cons in self._constraints.linear_constr['terminal_ineq']:
                GG = np.block([np.zeros((len(cons[0][:,0]),
                                         (self._nu + self._nx)*(self._N))),
                               mpc.convert(cons[0])])
                
                A = np.vstack((A,GG))
                self._l = np.hstack((self._l,
                                     mpc.convert(cons[1]).flatten()))
                self._u = np.hstack((self._u,
                                     np.ones(len(GG[:,0]))*np.inf))
            
            self._A = sparse.csc_matrix(A)
            
            P = np.block([[Q,NN],[NN.T,R]])
            P_ = np.copy(P)
            for i in range(self._N - 1):
                P_ = block_diag(P_, P)
            P_ = block_diag(P_, PP)
            self._P = sparse.csc_matrix(P_)
            
        else:
            QQ = np.copy(Q)
            RR = np.copy(R)
            N2 = np.copy(NN)
            for i in range(self._N - 1):
                QQ = block_diag(QQ, Q)
                RR = block_diag(RR, R)
                N2 = block_diag(N2, NN)
            N2 = np.vstack((N2, np.zeros((self._nx, self._nu*(self._N)))))
            QQ = block_diag(QQ, PP)
            
            A = mpc.convert(self._system.f[0], 'numpy')
            B = mpc.convert(self._system.f[1], 'numpy')
            
            G = np.vstack((np.zeros((self._nx, 1)), np.ones((self._nu, 1))))
            H = np.block([[np.eye(self._nx)],
                          [np.zeros((self._nu, self._nx))]])
            
            GG = np.copy(G)
            HH = np.copy(H)
            for i in range(self._N - 1):
                GG = block_diag(GG, G)
                HH = block_diag(HH, H)  
            HH = block_diag(HH, np.eye(self._nx))
            GG = np.block([[GG], [np.zeros((self._nx, self._N*self._nu))]])
            
            AA = np.eye(self._nx)
            BB = np.zeros((self._nx, self._nu*self._N))
            for i in range(self._N):
                AA = np.vstack((AA, matrix_power(A, (i + 1))))
                BBrow = np.copy(B)
                for j in range(i):
                    BBrow = np.block([matrix_power(A, (j + 1))@B, BBrow])
                BBrow = np.block([BBrow, 
                                  np.zeros((self._nx, 
                                            self._nu*(self._N - i - 1)))])
                BB = np.vstack((BB, BBrow))
                
            self._l = np.zeros(self._nx*(self._N + 1) + self._nu*self._N)
            self._u = np.zeros(self._nx*(self._N + 1) + self._nu*self._N)
            for i in range(self._N):
                li = (i + 1)*self._nx + i*self._nu
                ui = (i + 1)*self._nx+(i + 1)*self._nu
                self._l[li:ui] = mpc.convert(self._constraints.lower_bndu, 
                                             'numpy').flatten()
                li = i*self._nx + i*self._nu
                ui = (i + 1)*self._nx + i*self._nu
                self._l[li:ui] = mpc.convert(self._constraints.lower_bndx, 
                                             'numpy').flatten()
            for i in range(self._N):
                li = (i + 1)*self._nx + i*self._nu
                ui = (i + 1)*self._nx + (i + 1)*self._nu
                self._u[li:ui] = mpc.convert(self._constraints.upper_bndu, 
                                             'numpy').flatten()
                li = i*self._nx + i*self._nu
                ui = (i + 1)*self._nx + i*self._nu
                self._u[li:ui] = mpc.convert(self._constraints.upper_bndx, 
                                             'numpy').flatten()
            
            index = self._N*self._nx + self._nu*self._N
            self._l[index:] = mpc.convert(self._constraints.lower_bndend, 
                                          'numpy').flatten()
            self._u[index:] = mpc.convert(self._constraints.upper_bndend, 
                                          'numpy').flatten()
            
            
            for cons in self._constraints.linear_constr['ineq']:
                H = mpc.convert(cons[0], 'numpy')
                G = mpc.convert(cons[1], 'numpy')
                G1 = np.copy(G)
                H1 = np.copy(H)
                self._l = np.hstack((self._l, 
                                     mpc.convert(cons[2],'numpy').flatten()))
                self._u = np.hstack((self._u, 
                                     np.ones(len(cons[2]))*np.inf))
                for i in range(self._N - 1):
                    G1 = block_diag(G1, G)
                    H1 = block_diag(H1, H)
                    self._l = np.hstack((self._l, 
                                         mpc.convert(cons[2],'numpy').flatten()))
                    self._u = np.hstack((self._u, 
                                         np.ones(len(cons[2]))*np.inf))
                H1 = np.block([H1, np.zeros(len(H1[:,0], self._nx))])
                HH = np.vstack((HH, H1))
                GG = np.vstack((GG, G1))
            
            for cons in self._constraints.linear_constr['eq']:
                H = mpc.convert(cons[0], 'numpy')
                G = mpc.convert(cons[1], 'numpy')
                G1 = np.copy(G)
                H1 = np.copy(H)
                self._l = np.hstack((self._l, 
                                     mpc.convert(cons[2],'numpy').flatten()))
                self._u = np.hstack((self._u, 
                                     mpc.convert(cons[2],'numpy').flatten()))
                for i in range(self._N - 1):
                    G1 = block_diag(G1, G)
                    H1 = block_diag(H1, H)
                    self._l = np.hstack((self._l, 
                                         mpc.convert(cons[2], 
                                                     'numpy').flatten()))
                    self._u = np.hstack((self._u, 
                                         mpc.convert(cons[2], 
                                                     'numpy').flatten()))
                H1 = np.block([H1,np.zeros(len(H1[:,0], self._nx))])
                HH = np.vstack((HH ,H1))
                GG = np.vstack((GG, G1))
                
            for cons in self._constraints.linear_constr['terminal_ineq']:
                H1 = np.block([np.zeros((len(cons[1]),
                                         len(HH[0,:]) - self._nx)), 
                               mpc.convert(cons[0], 'numpy')])
                HH = np.block([[HH], [H1]])
                GG = np.block([[GG], [np.zeros((self._nx, self._N*self._nu))]])
                self._l = np.hstack((self._l, 
                                     mpc.convert(cons[1],'numpy').flatten()))
                self._u = np.hstack((self._u, 
                                     np.ones(len(cons[1]))*np.inf))
                
            for cons in self._constraints.linear_constr['terminal_eq']:
                H1 = np.block([np.zeros((len(cons[1]), 
                                         len(HH[0,:]) - self._nx)), 
                               mpc.convert(cons[0], 'numpy')])
                HH = np.block([[HH], [H1]])
                GG = np.block([[GG], 
                               [np.zeros((self._nx, self._N*self._nu))]])
                self._l = np.hstack((self._l, 
                                     mpc.convert(cons[1],'numpy').flatten()))
                self._u = np.hstack((self._u, 
                                     mpc.convert(cons[1],'numpy').flatten()))
                
            self._A = sparse.csc_matrix(GG + HH@BB)
            self._HH = HH
            self._AA = AA
            
            q = AA.T@(QQ@BB + N2)
            self._q = lambda x0 : x0.T@q
            self._P = sparse.csc_matrix(RR + BB.T@QQ@BB + BB.T@N2 + N2.T@BB)

        
        self._lqp = osqp.OSQP()
        
        rho = 0.1
        sigma =  1e-06
        max_iter = self._maxiter
        eps_abs = self._tol
        eps_rel = self._tol
        eps_prim_inf = 1e-04
        eps_dual_inf = 1e-04
        alpha = 1.6
        linsys_solver = 'qdldl'
        delta = 1e-06
        polish = False
        polish_refine_iter = 3
        verbose = self._verbose
        scaled_termination = False
        check_termination = 25
        warm_start = True
        scaling = 10
        adaptive_rho = True
        adaptive_rho_interval = 0
        adaptive_rho_tolerance = 5
        adaptive_rho_fraction = 0.4
        time_limit = 0
        
        for key in self._options.keys():
            locals()[key] = self._options[key]
        
        self._lqp.setup(P = self._P, A = self._A, l = self._l, u = self._u, 
                       rho = rho, sigma = sigma, max_iter = max_iter,
                       eps_abs = eps_abs, eps_rel = eps_rel, 
                       eps_prim_inf = eps_prim_inf, eps_dual_inf = eps_dual_inf,
                       alpha = alpha, linsys_solver = linsys_solver, delta = delta,
                       polish = polish, polish_refine_iter = polish_refine_iter,
                       verbose = verbose, scaled_termination = scaled_termination,
                       check_termination = check_termination, warm_start = warm_start,
                       scaling = scaling, adaptive_rho = adaptive_rho,
                       adaptive_rho_interval = adaptive_rho_interval,
                       adaptive_rho_tolerance = adaptive_rho_tolerance,
                       adaptive_rho_fraction = adaptive_rho_fraction,
                       time_limit = time_limit)
            
    def _init_casadi(self):
        
        self._optistack = cas.Opti() 
        
        self._U = self._optistack.variable(self._nu, self._N)      # for control trajectory   
        self._p = self._optistack.parameter(self._nx)             # parameter for initial value
        self._t = np.arange(self._t0, 
                            (self._t0 + (self._N+1)*self._system.h 
                             - (self._system.h/2)), 
                            self._system.h)
        
        if self._full_discretization:
            self._X = self._optistack.variable(self._nx, self._N + 1)     # for state trajectory
            # initial conditions
            self._optistack.subject_to(self._X[:,0]  == self._p)
            if self._system.autonomous:
                # add system dynamics as constrains
                for k in range(self._N):
                    x_next = mpc.convert(self._system.system_discrete(0, self._X[:,k], self._U[:,k]))
                    for i in range(self._nx):
                        self._optistack.subject_to(self._X[i, k + 1] == x_next[i])
                
        else:
            self._X = mpc.array((self._nx, self._N + 1))
            self._X[:,0] = self._p
            for k in range(self._N):
                t = self._t0 + k*self._system.h
                self._X[:,k + 1] = self._system.system_discrete(t, self._X[:,k], self._U[:,k])
        
        # Set objective
        J = self._objective.J(self._t, self._X, self._U, self._N)
        J = mpc.convert(J)
        self._optistack.minimize(J)
        
        # Set bounds
        for i in range(self._nx):
            for j in range(self._N):
                self._optistack.subject_to(
                    self._optistack.bounded(self._constraints.lower_bndx[i], 
                                            self._X[i, j], 
                                            self._constraints.upper_bndx[i])) 
        
        for i in range(self._nu):
            self._optistack.subject_to(
                self._optistack.bounded(self._constraints.lower_bndu[i], 
                                        self._U[i, :], 
                                        self._constraints.upper_bndu[i])) 
        
        for i in range(self._nx):
            self._optistack.subject_to(
                self._optistack.bounded(self._constraints.lower_bndend[i], 
                                        self._X[i,self._N], 
                                        self._constraints.upper_bndend[i]))
        
        #set linear constraints
        for cons in self._constraints.linear_constr['eq']:
            for k in range(self._N):
                cons_value = cons[0]@self._X[:, k] + cons[1]@self._U[:, k]
                for i in range(cons_value.dim[0]):
                    self._optistack.subject_to(cons_value[i] <= cons[2][i])
                    self._optistack.subject_to(cons_value[i] >= cons[2][i])
        for cons in self._constraints.linear_constr['ineq']:
            for k in range(self._N):
                cons_value = cons[0]@self._X[:, k] + cons[1]@self._U[:, k]
                for i in range(cons_value.dim[0]):
                    self._optistack.subject_to(cons_value[i] >= cons[2][i])
                
        for cons in self._constraints.linear_constr['terminal_eq']:
            cons_value = cons[0]@self._X[:, self._N]
            for i in range(cons_value.dim[0]):
                self._optistack.subject_to(cons_value[i] <= cons[1][i])
                self._optistack.subject_to(cons_value[i] >= cons[1][i])
        for cons in self._constraints.linear_constr['terminal_ineq']:
            cons_value = cons[0]@self._X[:, self._N]
            for i in range(cons_value.dim[0]):
                self._optistack.subject_to(cons_value[i] >= cons[1][i])
                
        #set nonlinear constraints
        for cons in self._constraints.nonlinear_constr['eq']:
            for k in range(self._N):
                cons_value = cons(self._t[k],self._X[:, k],self._U[:, k])
                for i in range(cons_value.dim[0]):
                    self._optistack.subject_to(cons_value[i] <= 0)
                    self._optistack.subject_to(cons_value[i] >= 0)
        for cons in self._constraints.nonlinear_constr['ineq']:
            for k in range(self._N):
                cons_value = cons(self._t[k],self._X[:, k],self._U[:, k])
                for i in range(cons_value.dim[0]):
                    self._optistack.subject_to(cons_value[i] >= 0)
                
        for cons in self._constraints.nonlinear_constr['terminal_eq']:
            cons_value = cons(self._t[-1],self._X[:, self._N])
            for i in range(cons_value.dim[0]):
                self._optistack.subject_to(cons_value[i] <= 0)
                self._optistack.subject_to(cons_value[i] >= 0)
        for cons in self._constraints.nonlinear_constr['terminal_ineq']:
            cons_value = cons(self._t[-1],self._X[:, self._N])
            for i in range(cons_value.dim[0]):
                self._optistack.subject_to(cons_value[i] >= 0)
                
        # set the optimizer
        options = {}
        options['max_iter'] = self._maxiter
        if self._method == 'ipopt':
            options['sb'] = 'yes'
            if self._verbose is False:
                options['print_level'] = 0
            options['acceptable_tol'] = self._tol
            options.update(self._options)
            self._optistack.solver("ipopt", 
                                   {'print_time': self._verbose},
                                   options)
        elif self._method == 'sqpmethod':
            if self._verbose is False:
                options['print_time'] = False
                options['print_header'] = False
            options['tol_du'] = self._tol
            options['tol_pr'] = self._tol
            options.update(self._options)
            self._optistack.solver('sqpmethod', options)
        
    def _init_scipy(self):
        self._bnds = ()
        lower_bndu = mpc.convert(self._constraints.lower_bndu, 'numpy')
        upper_bndu = mpc.convert(self._constraints.upper_bndu, 'numpy')
        for i in range(self._N):
            for j in range(self._nu):
                self._bnds += ((lower_bndu[j], upper_bndu[j]),)
                
        if self._full_discretization:
            self._Z_start = np.zeros(self._nu*self._N + self._nx*(self._N + 1))
            self._constraints_scipy = ({'type': 'eq', 
                                        'fun': self._cons_dynamics},)
            
            lower_bndx = mpc.convert(self._constraints.lower_bndx, 'numpy')
            upper_bndx = mpc.convert(self._constraints.upper_bndx, 'numpy')
            for i in range(self._N):
                for j in range(self._nx):
                    self._bnds += ((lower_bndx[j] ,upper_bndx[j]),)
            
            lower_bndend = mpc.convert(self._constraints.lower_bndend, 'numpy')
            upper_bndend = mpc.convert(self._constraints.upper_bndend, 'numpy')
            for j in range(self._nx):       
                self._bnds += ((lower_bndend[j], upper_bndend[j]),)
            
        else:
            self._Z_start = np.zeros(self._nu*self._N)
            self._constraints_scipy = ({'type': 'ineq', 
                                        'fun': self._bndsx_scipy},)
        if (self._constraints.nonlinear_constr['eq'] != [] or 
            self._constraints.linear_constr['eq'] != [] or 
            self._constraints.nonlinear_constr['terminal_eq'] != [] or 
            self._constraints.linear_constr['terminal_eq'] != []):
            self._constraints_scipy += ({'type': 'ineq', 
                                         'fun': self._cons_eq_scipy},)
        if (self._constraints.nonlinear_constr['ineq'] != [] or 
            self._constraints.linear_constr['ineq'] != [] or 
            self._constraints.nonlinear_constr['terminal_ineq'] != [] or 
            self._constraints.linear_constr['terminal_ineq'] != []):
            self._constraints_scipy += ({'type': 'ineq', 
                                         'fun': self._cons_ineq_scipy},)
                    
    def _J_scipy(self, Z):
        t = np.arange(self._t0, 
                      self._t0 + (self._N+1)*self._system.h, 
                      self._system.h)
        U = np.reshape(Z[:self._N*self._nu], (self._nu, self._N))
        if self._full_discretization:
            X = np.reshape(Z[self._N*self._nu:], (self._nx, self._N + 1))
        else:
            X = np.zeros((self._nx, self._N + 1))
            X[:, 0] = self._x0
            for k in range(self._N):
                X[:, k + 1] = mpc.convert(
                    self._system.system_discrete(t[k], X[:, k], U[:, k]),
                    'numpy').flatten()
        J = 0.
        if self._objective._type == 'NLP':
            for k in range(self._N):
                J += self._objective._L(t[k],X[:, k],U[:, k])
        else:
            for k in range(self._N):
                J += (X[:, k].T@self._objective._L[0].A@X[:, k]
                      + U[:, k].T@self._objective._L[1].A@U[:, k]
                      + X[:, k].T@self._objective._L[2].A@U[:, k]*2)
                
        if self._objective._F is not None:
            if self._objective._type == 'NLP':
                J += self._objective._F(t[:,self._N], X[:, self._N])
            else:
                J += X[:, self._N].T@self._objective._F.A@X[:, self._N]
        
        return J
    
    def _cons_dynamics(self, Z):
        U = np.reshape(Z[:self._N*self._nu], (self._nu,self._N))
        X = np.reshape(Z[self._N*self._nu:], (self._nx,self._N + 1))
        X_next = np.zeros((self._nx, self._N + 1))
        X_next[:, 0] = self._x0
        for k in range(self._N):
            t = self._t0 + k*self._system.h
            X_next[:, k + 1] = mpc.convert(
                self._system.system_discrete(t, X[:,k], U[:,k]),
                'numpy').flatten()
        return (X_next-X).flatten()
    
    def _cons_eq_scipy(self, Z):
        t = np.arange(self._t0, 
                      self._t0 + self._N*self._system.h, 
                      self._system.h)
        U = np.reshape(Z[:self._N*self._nu], (self._nu, self._N))
        if self._full_discretization:
            X = np.reshape(Z[self._N*self._nu:], (self._nx, self._N + 1))
        else:
            X = np.zeros((self._nx, self._N + 1))
            X[:, 0] = self._x0
            for k in range(self._N):
                X[:, k + 1] = mpc.convert(
                    self._system.system_discrete(t[k], X[:, k], U[:, k]),
                    'numpy').flatten()
        c = []
        
        for cons in self._constraints.linear_constr['eq']:
            for k in range(self._N):
                c += [cons[0]@X[:,k] + cons[1]@U[:, k] - cons[2]]
                c += [-c[-1]]
        
        for cons in self._constraints.linear_constr['terminal_eq']:
            c += [cons[0]@X[:, self._N] - cons[1]]
            c += [-c[-1]]
        
        for cons in self._constraints.nonlinear_constr['eq']:
            for k in range(self._N):
                c += [cons(t[k], X[:, k], U[:, k])]
                c += [-c[-1]]
                
        for cons in self._constraints.nonlinear_constr['terminal_eq']:
            c += [cons(t[-1],X[:,self._N])]
            c += [-c[-1]]
        
        c = flat_list(c)
        return np.array(c).flatten()
    
    def _cons_ineq_scipy(self, Z):
        t = np.arange(self._t0, 
                      self._t0 + self._N*self._system.h, 
                      self._system.h)
        U = np.reshape(Z[:self._N*self._nu], (self._nu, self._N))
        if self._full_discretization:
            X = np.reshape(Z[self._N*self._nu:], (self._nx, self._N + 1))
        else:
            X = np.zeros((self._nx, self._N + 1))
            X[:, 0] = self._x0
            for k in range(self._N):
                X[:, k + 1] = mpc.convert(
                    self._system.system_discrete(t[k], X[:,k], U[:,k]),
                    'numpy').flatten()
        c = []
        
        for cons in self._constraints.linear_constr['ineq']:
            for k in range(self._N):
                c += [cons[0]@X[:, k] + cons[1]@U[:, k] - cons[2]]
        
        for cons in self._constraints.linear_constr['terminal_ineq']:
            c += [cons[0]@X[:, self._N] - cons[1]]
        
        for cons in self._constraints.nonlinear_constr['ineq']:
            for k in range(self._N):
                c += [cons(t[k], X[:, k], U[:, k])]
                
        for cons in self._constraints.nonlinear_constr['terminal_ineq']:
            c += [cons(t[-1], X[:, self._N])]
        
        c = flat_list(c)
        return np.array(c).flatten()
        
    
    def _bndsx_scipy(self,Z):
        U = np.reshape(Z[:self._N*self._nu], (self._nu, self._N))
        X = np.zeros((self._nx, self._N + 1))
        X[:,0] = self._x0
        for k in range(self._N):
            t = self._t0 + k*self._system.h
            X[:, k + 1] = mpc.convert(
                self._system.system_discrete(t, X[:, k], U[:, k]),
                'numpy').flatten()
        c = np.zeros(2*self._nx*(self._N + 1))
        for i in range(self._N):
            for j in range(self._nx):
                if not self._constraints.upper_bndx[j] == mpc.inf:
                    c[i*self._nx + j] = self._constraints.upper_bndx[j] - X[j, i] 
                if not self._constraints.lower_bndx[j] == -mpc.inf:
                    c[self._nx*(self._N + 1) + i*self._nx + j] = -(self._constraints.lower_bndx[j] - X[j, i])
        for j in range(self._nx):
            if not self._constraints.upper_bndx[j] == mpc.inf:
                c[self._N*self._nx + j] = self._constraints.upper_bndend[j] - X[j, i] 
            if not self._constraints.lower_bndx[j] == -mpc.inf:
                c[self._nx*(self._N + 1)+ self._N*self._nx + j] = -(self._constraints.lower_bndend[j] - X[j, i])
        return c
            
    def solve(self, t, x0):
        """Start the optimization progress.

        Parameters
        ----------
        t : float
            Current time.
        x0 : array
            Current state.

        Returns
        -------
        array
            Optimal control sequence.
        array
            Optimal trajectory.

        """
        
        if not isinstance(t, (float,int)):
            raise TypeError('time t must be of type integer or float')
            
        if isinstance(x0, mpc.array):
            if x0.dim != (self._nx, 1):
                raise ValueError('x0 has the wrong dimension - ' + str(x0.dim) + ' != (' + str(self._nx) + ',1)')
        else: raise TypeError('x0 must be of type array - not ' + str(type(x0)))
        
        self._t0 = t
        self._X_start[:, 0] = x0
        self._x0 = mpc.convert(x0, 'numpy').flatten()
        
        #calculate the state fitting to the initial guess for the control (U_start)
        if self._full_discretization:
            for k in range(self._N):
                t = self._t0 + k*self._system.h
                self._X_start[:, k + 1] = self._system.system_discrete(
                    t, self._X_start[:, k], self._U_start[:, k])
                
        if self._solver == 'osqp':
            if self._full_discretization:
                self._l[:self._nx] = self._x0
                self._u[:self._nx] = self._x0
            
                self._lqp.update(l=self._l, u=self._u)
            else:
                q = self._q(self._x0)
                l = self._l - self._HH@self._AA@self._x0
                u = self._u - self._HH@self._AA@self._x0
                self._lqp.update(q=q, l=l, u=u)
            
            res = self._lqp.solve()
            if res.info.status_val != 1:
                return (None, res.info.status)
            
            UU = mpc.array((self._nu,self._N))
            XX = mpc.array((self._nx,self._N+1))
            
            for i in range(self._N):
                if self._full_discretization:
                    UU[:, i] = res.x[(i + 1)*self._nx + i*self._nu:
                                    (i + 1)*self._nx + (i + 1)*self._nu]
                    XX[:, i] = res.x[i*self._nx+i*self._nu:
                                    (i + 1)*self._nx+i*self._nu]
                else:
                    UU[:, i] = res.x[i*self._nu:(i + 1)*self._nu]
                    XX[:, 0] = self._x0
                    for k in range(self._N):
                        t = self._t0 + (k + 1)*self._system.h
                        XX[:, k + 1] = self._system.system_discrete(
                            t, XX[:, k], UU[:, k])
          
        elif self._solver == 'casadi':
            t = np.arange(self._t0, 
                          (self._t0 
                           + (self._N+1)*self._system.h 
                           - (self._system.h/2)), 
                          self._system.h)
            
            if self._system.autonomous:
                # set x0 for optimization
                self._optistack.set_value(self._p, self._x0)
            else:
                self._init_casadi()
                # set x0 for optimization
                self._optistack.set_value(self._p, self._x0)
                # add system dynamics as constrains
                for k in range(self._N):
                    x_next = self._system.system_discrete(
                        t[k], self._X[:, k], self._U[:, k])
                    for i in range(self._nx):
                        self._optistack.subject_to(
                            self._X[i, k + 1] == x_next[i])
                        
            # set timespan
            self._optistack.set_value(self._t, t)
            # set initial guess for optimization 
            self._optistack.set_initial(self._U, mpc.convert(self._U_start.A,'numpy')) 
            if self._full_discretization:
                self._optistack.set_initial(self._X, mpc.convert(self._X_start.A,'numpy')) 
            
            try:
                sol = self._optistack.solve() # solve ocp
            except Exception as error:
                return (None,error)
            
            # save control from the current solution as initial guess for the next interation step
            U = np.reshape(sol.value(self._U), (self._nu, self._N))
            if self._full_discretization:
                X = np.reshape(sol.value(self._X), (self._nx, self._N + 1))
            else:
                for k in range(self._N):
                    if self._system.autonomous:
                        t = 0.
                    else:
                        t = self._t0 + k*self._system.h
                    self._X_start[:, k + 1] = self._system.system_discrete(
                        t, self._X_start[:, k], U[:, k])
                X = self._X_start
                
            #self._U_start.copy(U)
            self.U_start = mpc.array(U)
            
            UU = mpc.array(U)
            XX = mpc.array(X)
            
        elif self._solver == 'scipy':
            
            self._Z_start[:self._N*self._nu] = mpc.convert(
                self._U_start, 'numpy').flatten()
            
            if self._full_discretization:
                self._Z_start[self._N*self._nu:] = mpc.convert(
                    self._X_start, 'numpy').flatten()
                
            options = {}
            options['maxiter'] = self._maxiter
            options['disp'] = self._verbose
            options.update(self._options)
            
            res = minimize(self._J_scipy,self._Z_start, 
                           constraints=self._constraints_scipy, 
                           bounds=self._bnds, tol=self._tol, 
                           method=self._method, options=self._options)
            
            if res.success == False: 
                error = res.message
                return (None, error)
            
            U = np.reshape(res.x[:self._nu*self._N], 
                           (self._nu, self._N))
            if self._full_discretization:
                X = np.reshape(res.x[self._nu*self._N:], 
                               (self._nx, self._N + 1))
            else:
                for k in range(self._N):
                    t = self._t0+ k*self._system.h
                    self._X_start[:, k + 1] = self._system.system_discrete(
                        t, self._X_start[:, k], U[:, k])
                X = self._X_start
                
            self._U_start = mpc.array(U)
            
            UU = mpc.array(U)
            XX = mpc.array(X)
            
        return (UU, XX)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: Jonas Schiessl


import casadi as cas

from scipy.integrate import solve_ivp

from inspect import signature
import dill

import nmpyc as mpc
from nmpyc.utils import mpc_convert

class system:
    """
    A class used to define the system dynamics of an optimal control problem.

    The dynamics can be discrete or continuous. 
    A discrete system is defined by a difference equation 

    .. math::

       x(t_{k+1}) = f(t_k,x(t_k),u(t_k))

    and a continous system is defined by the ordinary differential equation

    .. math::

       \dot{x}(t_k)=f(t_k,x(t_k),u(t_k))).

    In the letter case the differential equation will be discretized by a choosen 
    integration method.

    Parameters
    ----------
    f : callable
        Function defining the right hand side of the system dynamics of 
        the form :math:`f(t,x,u)` or :math:`f(x,u)` in the 
        :py:attr:`~autonomous` case. See also :py:attr:`~f`.
    nx : int
        Dimension of the state. Must be a positive integer. 
        See also :py:attr:`~nx`.
    nu : int
        Dimension of the control. Must be a positive integer. 
        See also :py:attr:`~nu`.
    system_type : str, optional
        String defining if the given system dynamics are 
        discrete or continuous. The default is 'discrete'.
    sampling_rate : float, optional
        Sampling rate defining at which time instances the 
        dynamics are evaluated. The default is 1.
    t0 : float, optional
        Initial time for the optimal control problem. The default is 0.
        See also :py:attr:`~t0`.
    method : str, optional
        String defining which integration method should be used to discretize 
        the system dynamics. The default is 'cvodes'. 
        For further informations about the provided integrators see 
        :py:attr:`~method`.
        
    """
    
    def __init__(self, f, nx, nu, system_type='discrete', 
                 sampling_rate=1., t0=0., method='cvodes'):
        
        self._options = {}
        self._type = 'NLP'
        
        self.f = f
        self.system_type = system_type
        self.h = sampling_rate
        self.t0 = t0
        self.nx = nx
        self.nu = nu
        self.method = method
            
        if self._type == 'LQP' and self._system_type == 'continuous':
            if self._method in ['rk4', 'euler', 'heun']:
                self._A = self._f[0]
                self._B = self._f[1]
                self._discretizeLQP()
            else:
                A = self._f[0]
                B = self._f[1]
                self._f = lambda x,u : A@x + B@u
                self._type = 'NLP'
                
    @property 
    def f(self):
        """callable or list of array: Right hand side :math:`f(t,x,u)` of the system dynamics.
        
        The return value of this attribute depends on how the system is initialized.
        If it is initialized as a linear system by :py:meth:`~LQP` a list containing the arrays defining the 
        system dynamics are returned.
        If the system is initalized by a possible nonlinear callable function this function is returned. 
        Note, that even if :py:attr:`~autonomous` is True the returned funtion depends on the time 
        and always has the form :math:`f(t,x,u)`.
        """
        return self._f
    
    @f.setter 
    def f(self, ff):
        if callable(ff):
            sig = signature(ff)
            params = sig.parameters
            if len(params) == 2:
                self._autonomous = True
                F = lambda t,x,u : ff(x, u)
            elif len(params) == 3:
                F = ff
                self._autonomous = False
            else:
                raise ValueError(
                    'f must have two or three input arguments - not ' 
                    + str(len(params)))
        elif isinstance(ff, list):
            for i in range(2):
                if not isinstance(ff[i], mpc.array): 
                    raise TypeError(
                        'f must be a list of array -  not ' 
                        + str(type(ff[i])))
            self._autonomous = True
            self._type = 'LQP'
            F = ff
        else:
            raise TypeError(
                'f must be callable or list of arrays - not ' + str(type(ff)))
        self._f = F
        
    @property 
    def system_type(self):
        """str : String defining whether the dynamics are discrete or continuous.
        
        A discrete system is defined by a difference equation 

        .. math::

           x(t_{k+1}) = f(t_k,x(t_k),u(t_k))

        and a continous system is defined by the ordinary differential equation

        .. math::

           \dot{x}(t_k)=f(t_k,x(t_k),u(t_k))).

        """
        return self._system_type
    
    @system_type.setter 
    def system_type(self, syst_type):
        if not isinstance(syst_type, str): 
            raise TypeError(
                'system type must be a string - not ' + str(type(syst_type)))
        if syst_type not in ['discrete', 'continuous']:
            raise ValueError(
                'system_type must be discrete or continuous - not ' 
                + syst_type)
        self._system_type = syst_type
        
    @property 
    def h(self):
        """float : Sampling time :math:`h` of the system.
        
        This attribute defines at which time instances the 
        dynamics are evaluated. 
        This means the time :math:`t_k` is given by the equation

        .. math::
           
           t_k = t_0 + kh.

        In addition, the control values are assumed to be constant during a sampling instance
        and can only be change at the times :math:`t_k`.
        """
        return self._h
    
    @h.setter 
    def h(self, sampling_rate):
        if not isinstance(sampling_rate, (float, int)):
            raise TypeError(
                'sampling_rate must be of type integer or float - not ' 
                + str(type(sampling_rate)))
        if sampling_rate > 0:
            self._h = sampling_rate
        else: 
            raise ValueError(
                'sampling rate must be greter than zero - not ' 
                + str(sampling_rate))
        
    @property 
    def t0(self):
        """float : Initial time of the optimal control problem.
        
        The initial state :math:`x_0` is measured 
        at time :math:`t_0`.
        The state :math:`x(t)` is evaluated at the time instances 
        :math:`t_0 + kh` during the MPC loop where :math:`h` is 
        the :py:attr:`~sampling_rate`.
        """
        return self._t0
    
    @t0.setter 
    def t0(self, t_0):
        if not isinstance(t_0, (float, int)):
            raise TypeError(
                't0 must be of type integer or float - not ' + str(type(t_0)))
        self._t0 = t_0
        
    @property 
    def nx(self):
        """int : Dimension of the state. 
        
        The value of :math:`x(t)` at a given time :math:`t_k` 
        is a element of :math:`\mathbb{R}^{nx}`.
        In the linear case this value equals with the dimension 
        of the system matrix :math:`A \in \mathbb{R}^{nx \\times nx}`.
        """
        return self._nx
    
    @nx.setter 
    def nx(self, n_x):
        if not isinstance(n_x, int):
            raise TypeError(
                'nx must be of type integer - not ' + str(type(n_x)))
        if n_x > 0:
            self._nx = n_x
        else: 
            raise ValueError('nx must be grater than zero - not ' + str(n_x))
        
    @property 
    def nu(self):
        """int : Dimension of the control.
        
        The value of :math:`u(t)` at a given time :math:`t_k` 
        is a element of :math:`\mathbb{R}^{nu}`.
        In the linear case this value equals with the dimension 
        of the columns of the control matrix :math:`B \in \mathbb{R}^{nx \\times nu}`.
        """
        return self._nu
    
    @nu.setter 
    def nu(self, n_u):
        if not isinstance(n_u, int):
            raise TypeError(
                'nu must be of type integer - not ' + str(type(n_u)))
        if n_u > 0:
            self._nu = n_u
        else: 
            raise ValueError('nu must be grater than zero - not ' + str(n_u))
        
    @property 
    def method(self):
        """str : Integration method for discretization of the dynamics.
        
        The following integrators are currently supported:   

        - from `CasADi <http://casadi.sourceforge.net/api/html/db/d3d/classcasadi_1_1Integrator.html>`_: `cvodes`, `idas`, `collocation`, `oldcollocation` and `rk`   

        - from `SciPy <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html>`_: `RK45`, `RK23`, `DOP853`, `Radau`, `BDF` and `LSODA` 
        
        - from nMPyC: `rk4`, `euler` und `heun` (fixed step integration methods)"""

        return self._method
    
    @method.setter
    def method(self, meth):
        if not isinstance(meth, str): 
            raise TypeError(
                'method must be of type string - not ' + str(type(meth)))
        self._method = meth
        if self._method in ['cvodes', 'idas', 
                            'collocation', 'oldcollocation', 'rk']:
            self._integrator = 'casadi'
        elif self._method in ['RK45', 'RK23', 'DOP853', 
                              'Radau', 'BDF', 'LSODA']:
            self._integrator = 'scipy'
        elif self._method in ['rk4', 'euler', 'heun']:
            self._integrator = 'mpc'
            self._options['number_of_finit_elements'] = 1
        else:
            raise ValueError(
                'allowed methods are cvodes, idas, collocation, ' 
                + 'oldcollocation, rk, RK45, RK23, DOP853, Radau, BDF, '
                + 'LSODA, euler, heun and rk4 - not ' + meth)
    
    @property 
    def autonomous(self):
        """bool : If True, the system is time-invariant. 
        
        The right hand side of the dynamics :math:`f(t,x,u)` 
        are not explicitly dependent on the time variable :math:`t`.
        In this case :math:`f(t,x,u)=f(x,u)` holds."""
        return self._autonomous
    
    @property
    def type(self):
        """str : Indicating whether the system dynamics are linear.
        
        If `LQP`, the system dynamics are linear.
        The right hand side of the system dynamics 
        is given by

        .. math::

           f(x,u) = Ax + Bu.

        It also implies that the system is :py:attr:`~autonomous`.

        If the system dynamics are not initialized as linear with the :py:meth:`~LQP` 
        method this attribute has the value `NLP`.
        """
        return self._type
    
    def __str__(self):
        string = ''
        
        string += 'autonomous: ' + str(self._autonomous) + '\n'
        string += 'system type: ' + str(self._system_type) + '\n'
        
        lqp = False
        if self.type == 'LQP': lqp = True
        string += 'linear: ' + str(lqp) + '\n'
        string += 'sampling rate: ' + str(self._h)
        
        if self.system_type == 'continuous': 
            string += '\n'
            string += 'integrator: ' + str(self.method)
        
        return string 
    
    @classmethod
    def LQP(cls, A, B , nx, nu, system_type='discrete', 
            sampling_rate=1., t0=0., method='euler'):
        """Initialize the system with linear dynamics.
        
        In this case the right hand side of the dynamics has the 
        form :
        
        .. math::
            
           f(x,u) = Ax+Bu 
            
        which is always :py:attr:`~autonomous`.   
        If not a fixed step method is choosen for integration the optimizer 
        can not use the linear structure of the problem during the 
        optimization process.
        
        Parameters
        ----------
        A : array
            Matrix definig the linear state input on the right hand side 
            of the dynamics.
        B : array
            Matrix definig the linear state input on the right hand side 
            of the dynamics.
        nx : int
            Dimension of the state. Must be a positive integer. 
            See also :py:attr:`~nx`.
        nu : int
            Dimension of the control. Must be a positive integer. 
            See also :py:attr:`~nu`.
        system_type : str, optional
            String defining whether the given system dynamics are 
            discrete or continuous. The default is 'discrete'.
        sampling_rate : float, optional
            Sampling rate defining at which time instances the 
            dynamics are evaluated. The default is 1.
        t0 : float, optional
            Initial time for the optimal control problem. The default is 0.
            See also :py:attr:`~t0`.
        method : str, optional
            String defining which integration method should be used to discretize 
            the system dynamics. The default is 'euler'. 
            For further informations about the provided integrators see 
            :py:attr:`~method`.

        Returns
        -------
        lqp : system
            nMPyC-system class object suitable to define a linear 
            quadratic optimal control problem..

        """
        
        if not isinstance(A, mpc.array):
            raise TypeError('A must be of type array - not ' + str(type(A)))
        if not A.dim == (nx,nx):
            raise ValueError(
                'A has the wrong dimension - ' 
                + str(A.dim) + '!=' + str((nx,nx)))
            
        if not isinstance(B, mpc.array):
            raise TypeError('B must be of type array - not ' + str(type(B)))
        if not B.dim == (nx,nu):
            raise ValueError(
                'B has the wrong dimension - ' 
                + str(B.dim) + '!=' + str((nx,nu)))
        
        if A.symbolic:
            raise ValueError(
                'A has to be purely numeric, but also has symbolic entries')   
        if B.symbolic:
            raise ValueError(
                'B has to be purely numeric, but also has symbolic entries')
        
        lqp = cls([A,B], nx, nu, system_type, sampling_rate, t0, method)
        
        return lqp
    
    def set_integratorOptions(self, options):
        """Set options for the integration method.

        Parameters
        ----------
        options : dict
            Dictionary containing the keywords of the required options 
            and their values.


        The available options are depending on the choosen :py:attr:`~method` of integration.
        For the nMPyC integrators the only available option is `number_of_finit_elements` 
        which must be an int greater than zero and defines how many discretation steps are 
        computed during one sampling period defined by the sampling rate.
        The available options for the CasADi integrators can be found at `Sourceforge <http://casadi.sourceforge.net/api/html/db/d3d/classcasadi_1_1Integrator.html>`_ and for the SciPy integrators at the `Scipy documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html>`_.

        """
        
        if not isinstance(options, dict):
            raise TypeError(
                'options must be of type dictionary - not ' 
                + str(type(options)))
        self._options = options
        
        if self._type == 'LQP' and self._system_type == 'continuous':
            if self._method in ['rk4', 'euler', 'heun']:
                self._f[0] = self._A 
                self._f[1] = self._B
                self._discretizeLQP()
    
    @mpc_convert
    def system_discrete(self, t, x, u):
        """Evaluate discretized right hand side of the system dynamics.

        Parameters
        ----------
        t : float
            Time instant at which the system dynamics are evaluated.
        x : array
            State value at which the system dynamics are evaluated.
        u : array
            Control value at which the system dynamics are evaluated.

        Returns
        -------
        array
            Value of the discretized right hand side of the dynamics 
            evaluated at the given inputs.

        """
        
        if self._system_type == 'discrete':
            return self.system(t, x, u)
        
        else:
            if self._type == 'LQP':
                return self._f_linear(x, u)
            
            if self._integrator == 'casadi':
                return self._solve_ode_casadi(t, x, u)
            
            if self._integrator == 'scipy':
                return self._solve_ode_scipy(t, x, u)
            
            if self._integrator == 'mpc':
                if self._method == 'rk4':
                    return self._solve_ode_rk4(t, x, u)
                elif self._method == 'euler':
                    return self._solve_ode_euler(t, x, u)
                elif self._method == 'heun':
                    return self._solve_ode_heun(t, x, u)
    
    @mpc_convert
    def _f_linear(self, x, u):
        y = self._f[0]@x + self._f[1]@u
        return y
    
    @mpc_convert
    def system(self, t, x, u):
        """Evaluate right hand side :math:`f(t,x,u)` of the dynamics.

        Parameters
        ----------
        t : float
            Time instant at which the system dynamics are evaluated.
        x : array
            State value at which the system dynamics are evaluated.
        u : array
            Control value at which the system dynamics are evaluated.

        Returns
        -------
        array
            Value of the possible not discrete right hand side of the dynamics 
            evaluated at the given inputs.

        """
        
        if self.type == 'NLP':
            return self._f(t,x,u)
        else:
            return self._f[0]@x + self._f[1]@u
        
    def _solve_ode_casadi(self, t, x, u):
        
        # Define symbolic variables for writing down the differential equation
        xx = cas.MX.sym('x', self._nx)    
        uu = cas.MX.sym('u', self._nu)   

        # Declare the ODE in casadi style
        ode = {}             
        ode['x'] = xx       
        ode['p'] = uu       
        
        tt = cas.MX.sym('t')
        xdot = self.system(tt, xx, uu)._A
        ode['t'] = tt
        
        ode['ode'] = xdot
        
        # Set the integrator options
        integr_opts = {}
        integr_opts.update(self._options)
        integr_opts['t0'] = t
        integr_opts['tf'] = t + self._h

        # Construct a Function to solve the ODE over a timestep dt
        F = cas.integrator('F', self._method, ode, integr_opts)

        # Solve the ODE
        x0 = mpc.convert(x)
        u = mpc.convert(u)
        res = F(x0=x0, p=u)
        y = mpc.array(res["xf"])
        
        return y
    
    def _solve_ode_scipy(self, t, x, u):
        
        def f(t,z): return mpc.convert(self.system(t,z,u), 'numpy').flatten() 
        
        z0 = mpc.convert(x,'numpy').flatten()
        sol = solve_ivp(f, [t,t+self._h], z0, method=self._method)
        y = mpc.array(sol.y[:,-1])
        
        return y
        
    def _solve_ode_rk4(self, t, x, u):
        
        steps = self._options['number_of_finit_elements']   # number of steps 
        tk = t
        dt = self._h/steps               # time step = sampling rate/steps
        x_next = mpc.array(x)           # copy x
        # Runge-Kutta 4 integration
        for i in range(steps):
            k1 = self.system(tk, x_next, u)
            k2 = self.system(tk + dt/2, x_next + k1*(dt/2), u)
            k3 = self.system(tk + dt/2, x_next + k2*(dt/2), u)
            k4 = self.system(tk + dt, x_next + k3*dt, u)
                
            x_next = x_next + (k1 + k2*2 + k3*2 + k4)*(dt/6)
            tk += dt
            
        return x_next
    
    def _solve_ode_euler(self, t, x, u):
        
        steps = self._options['number_of_finit_elements']
        tk = t
        dt = self.h/steps               
        x_next = mpc.array(x)           
        for i in range(steps):
            x_next = x_next + self.system(tk, x_next, u) * dt
            tk += dt
            
        return x_next
    
    def _solve_ode_heun(self, t, x, u):
        
        steps = self._options['number_of_finit_elements']   
        tk = t
        dt = self._h/steps               
        x_next = mpc.array(x)           
        for i in range(steps):
            xx = x_next + self.system(tk, x_next, u) * dt
            x_next = (x_next + (xx + self.system(tk + dt, xx, u)*dt))*0.5
            tk += dt
            
        return x_next
    
    def _discretizeLQP(self):
        
        if self._method == 'euler':
            
            steps = self._options['number_of_finit_elements']
            dt = self.h/steps 
            
            A_euler = self._f[0]*dt
            B_euler = self._f[1]*dt 
              
            A = mpc.eye(self._nx)   
            B = mpc.zeros((self._nx,self._nu))
            for i in range(steps):
                B = B + A@B_euler
                A = A + A@A_euler     
                
            self.f[0] = A
            self.f[1] = B
                
            
        if self._method == 'heun':
            
            steps = self._options['number_of_finit_elements']
            dt = self.h/steps
            
            A_k1 = self._f[0]
            B_k1 = self._f[1]
            
            A_k2 = self._f[0] + self._f[0]*dt@ A_k1
            B_k2 = self._f[0]*dt @ B_k1 + self._f[1]
            
            A_heun = (A_k1 + A_k2)*(dt/2)
            B_heun = (B_k1 + B_k2)*(dt/2)
            
            A = mpc.eye(self._nx)   
            B = mpc.zeros((self._nx,self._nu))
            for i in range(steps):
                B = B + A@B_heun
                A = A + A@A_heun     
                
            self.f[0] = A
            self.f[1] = B
            
        elif self.method == 'rk4':
            
            steps = self._options['number_of_finit_elements']
            dt = self.h/steps
            
            A_k1 = self._f[0]
            B_k1 = self._f[1]
            
            A_k2 = self._f[0] + self._f[0]@A_k1*(dt/2)
            B_k2 = self._f[0]@B_k1*(dt/2) + self._f[1]
            
            A_k3 = self._f[0] + self._f[0] @ A_k2*(dt/2)
            B_k3 = self._f[0]@B_k2*(dt/2) + self._f[1]
            
            A_k4 = self._f[0] + self._f[0] @ A_k3 * dt
            B_k4 = self._f[0]@B_k3*dt + self._f[1]
            
            A_rk4 = (A_k1 + A_k2*2 + A_k3*2 + A_k4)*(dt/6)
            B_rk4 = (B_k1 + B_k2*2 + B_k3*2 + B_k4)*(dt/6)
              
            A = mpc.eye(self._nx)   
            B = mpc.zeros((self._nx,self._nu))
            for i in range(steps):
                B = B + A@B_rk4
                A = A + A@A_rk4   
                
            self.f[0] = A
            self.f[1] = B
            
    
    def save(self, path):
        """Saving the system to a given file with `dill <https://dill.readthedocs.io/en/latest/dill.html>`_.
        
        The path can be absolut or relative and 
        the ending of the file is arbitrary.

        Parameters
        ----------
        path : str
            String defining the path to the desired file. 



        For example

        >>> system.save('system.pickle')
        
        will create a file `system.pickle` containing the nMPyC system object.
        """
        
        with open(path, "wb") as output_file:
            dill.dump(self, output_file, -1)
    
    @classmethod
    def load(cls, path):
        """Loads a nMPyC system object from a file.
        
        The specified path must lead to a file that was previously saved with
        :py:meth:`~save`.
        
        Parameters
        ----------
        path : str
            String defining the path to the file containing the nMPyC 
            system object. 
            
            
        For example

        >>> system.load('system.pickle')
        
        will load the system previously saved with :py:meth:`~save`.
            
        """
        
        try:
            with open(path, "rb") as input_file:
                e = dill.load(input_file)
        except:
            raise Exception(
                'Can not load system from file. File not readable!')
            
        if not isinstance(e, system):
            raise Exception(
                'Can not load system from file. File does not cotain a system!')
            
        return e



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: Jonas Schiessl

import inspect
from functools import wraps

import numpy as np

import nmpyc as mpc

def functionToString(func):
    str_func = inspect.getsource(func)
    name = func.__name__
    return (name, str_func)

def stringToFunction(name,str_func):
    if name == '<lambda>':
        index = str_func.find('lambda')
        func = eval(str_func[index:])
    else:
        exec(str_func)
        func = locals()[name]
    return func

def flat_list(t):
    flat_list = []
    for sub in t:
        for i in range(len(sub)):
            flat_list.append(sub[i])
            
    return flat_list

# def mpc_convert(func):
#     sig = inspect.signature(func)
#     params = sig.parameters
#     if inspect.getargspec(func)[0][0] == 'self':
#         if len(params) == 4:
#             def inner(self,t,x,u):
#                 if np.isscalar(x):
#                     x = [[x]]
#                 if np.isscalar(u):
#                     u = [[u]]
#                 x = mpc.array(x)
#                 u = mpc.array(u)
#                 y = func(self,t,x,u)
#                 if np.isscalar(y):
#                     y = [[y]]
#                 return mpc.array(y)
#         elif len(params) == 3:
#             def inner(self,t,x):
#                 if np.isscalar(x):
#                     x = [[x]]
#                 x = mpc.array(x)
#                 y = func(self,t,x)
#                 if np.isscalar(y):
#                     y = [[y]]
#                 return mpc.array(y)
#     else:
#         if len(params) == 3:
#             def inner(t,x,u):
#                 if np.isscalar(x):
#                     x = [[x]]
#                 if np.isscalar(u):
#                     u = [[u]]
#                 x = mpc.array(x)
#                 u = mpc.array(u)
#                 y = func(t,x,u)
#                 if np.isscalar(y):
#                     y = [[y]]
#                 return mpc.array(y)
#         elif len(params) == 2:
#             def inner(t,x):
#                 if np.isscalar(x):
#                     x = [[x]]
#                 x = mpc.array(x)
#                 y = func(t,x)
#                 if np.isscalar(y):
#                     y = [[y]]
#                 return mpc.array(y)
#     return inner

def mpc_convert(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        sig = inspect.signature(func)
        params = sig.parameters
        if inspect.getargspec(func)[0][0] == 'self':
            if len(params) == 4:
                def inner(self,t,x,u):
                    if np.isscalar(x):
                        x = [[x]]
                    if np.isscalar(u):
                        u = [[u]]
                    x = mpc.array(x)
                    u = mpc.array(u)
                    y = func(self,t,x,u)
                    if np.isscalar(y):
                        y = [[y]]
                    return mpc.array(y)
            elif len(params) == 3:
                def inner(self,t,x):
                    if np.isscalar(x):
                        x = [[x]]
                    x = mpc.array(x)
                    y = func(self,t,x)
                    if np.isscalar(y):
                        y = [[y]]
                    return mpc.array(y)
        else:
            if len(params) == 3:
                def inner(t,x,u):
                    if np.isscalar(x):
                        x = [[x]]
                    if np.isscalar(u):
                        u = [[u]]
                    x = mpc.array(x)
                    u = mpc.array(u)
                    y = func(t,x,u)
                    if np.isscalar(y):
                        y = [[y]]
                    return mpc.array(y)
            elif len(params) == 2:
                def inner(t,x):
                    if np.isscalar(x):
                        x = [[x]]
                    x = mpc.array(x)
                    y = func(t,x)
                    if np.isscalar(y):
                        y = [[y]]
                    return mpc.array(y)
        return inner(*args, **kwargs)
    return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Author: Jonas Schiessl

"""
Module for array definition and computation.

This module provides an array class and associated functions for 
corresponding matrix calculations.

The goal of this class and the individual functions is to enable 
compatibility of calculations with both casadi and numpy objects 
without changing the syntax of the program. 
This enables the user to program as easily as possible and at 
the same time to switch between symbolic and numeric calculation.
"""
import numpy as np
from numpy import linalg as LA
import casadi as cas

# Constants
inf = np.inf
"""float : Constant to define infinity."""
pi = np.pi
"""float : Constant to define the number pi."""

class array:
    """
    Class used to save arrays with symbolic or numeric values. 
    
    The symbolic entries are provided by CasADi and will be 
    transformed automatically to numeric values of numpy type 
    if it is posiible.
    
    Parameters
    ---------
    dim : int, tuple, cas.MX, cas.SX, cas.DM, list or numpy.ndarray, optional
        Dimension of which an empty array is created or object from which 
        the entries and dimension are copied. The default is 0.

    """
    
    def __init__(self, dim=0):
        
        if isinstance(dim, int):
            self._dim = (dim, 1)
        elif isinstance(dim, float):
            if (dim).is_integer():
                self._dim = (int(dim), 1)
            else: 
                raise ValueError('dimensions must be integers')
        elif isinstance(dim, tuple):
            if (not isinstance(dim[0], (int, float)) or not 
                isinstance(dim[1], (int, float))):
                raise ValueError('dimension must be tupel of integers')
            if isinstance(dim[0], float) and not (dim[0]).is_integer():
                raise ValueError('dimension must be tupel of integers')
            if isinstance(dim[1], float) and not (dim[1]).is_integer():
                raise ValueError('dimension must be tupel of integers')
            self._dim = dim
        else:
            self._dim = (0,0)
            
        if self._dim[0]<0 or self._dim[1]<0:
            raise ValueError('negativ dimensions are not allowed')
            
        self._A = np.zeros(self._dim)
        self._casadi = False
        
        if isinstance(dim, (np.ndarray, cas.MX, cas.SX, cas.DM, array, list)):
            self._copy(dim)
        else: 
            if not isinstance(dim, (int, float, tuple)):
                raise TypeError(
                    'input parameter must be of type' + 
                    ' int, tuple, list, numpy.ndarray,' + 
                    ' casadi.MX, casadi.SX or casadi.DX - not ' 
                    + str(type(dim)))
                
        if not len(self._dim) == 2:
            raise ValueError('only two dimesnions are allowed')
            
    @property 
    def A(self):
        """ casadi.MX or numpy.array : Array containing all entries."""
        try:
            AA = np.array(cas.evalf(self._A))
            return AA
        except: 
            return self._A
    
    @property 
    def dim(self):
        """tuple : Dimension of the array."""
        return self._dim
    
    @property
    def symbolic(self):
        """bool : True if array has symbolic entries, False otherwise."""
        return self._casadi
    
    @property
    def T(self):
        """array : Transposed array."""
        return self.transpose()
        
    def __str__(self): 
        return str(self._A)
    
    def __len__(self): 
        if self._casadi:
            return self._A.shape()[1]
        else:
            return len(self._A)
    
    def __neg__(self): 
        C = array(-self._A)
        return C
    
    def __pos__(self):
        C = array(-self._A)
        return C
    
    def __abs__(self): 
        return abs(self)
    
    def __int__(self): 
        C = int(self._A)
        return C
    
    def __float__(self): 
        C = float(self._A)
        return C
    
    def __add__(self, other): 
    
        if not isinstance(other, 
                          (array, np.ndarray, cas.MX, cas.SX, cas.DM, 
                           int, float)):
             raise TypeError(
                 'input parameter must be of type' + 
                 ' array, int, float, numpy.ndarray,' + 
                 ' casadi.MX, casadi.SX or casadi.DX - not ' 
                 + str(type(other)))
        
        C = array(self._dim)
        
        if isinstance(other, array):
            if self._dim != other._dim:
                raise ValueError(
                    'arrays must have the same dimensions. ' 
                    + str(self._dim) + 
                    ' != ' + str(other._dim))
            other = other._A
        
        if isinstance(other, np.ndarray) and len(other.shape) == 1:
            if self._dim[0] == 1:
                other = np.reshape(other,(self._dim[0],other.shape[0]))
            elif self._dim[1] == 1:
                other = np.reshape(other,(other.shape[0],self._dim[1]))
            else: 
                raise ValueError(
                    'arrays must have the same dimensions. ' 
                    + str(self._dim) + ' != ' + str(other.shape))
             
        C._A = self._A + other
        
        try:
            C._A = np.array(cas.evalf(C._A))
        except: 
            C._casadi = True
        
        return C
    
    def __radd__(self, other): 
        return self.__add__(other)
    
    def __sub__(self, other):
        return(self.__add__((-other)))
    
    def __rsub__(self, other): 
        return -(self.__sub__(other))
        
    def __mul__(self,other): 
    
        if not isinstance(other, 
                          (array, np.ndarray, cas.MX, cas.SX, cas.DM, 
                           int, float)):
             raise TypeError(
                 'input parameter must be of type' 
                 + ' array, int, float, numpy.ndarray,' + 
                 ' casadi.MX, casadi.SX or casadi.DX - not ' 
                 + str(type(other)))
        
        C = array(self._dim)
        
        if isinstance(other, np.ndarray) and len(other.shape) == 1:
            if self._dim[0] == 1:
                other = np.reshape(other,(1, other.shape[0]))
            elif self._dim[1] == 1:
                other = np.reshape(other, (other.shape[0], 1))
            else: 
                raise ValueError(
                    'arrays must have the same dimensions. ' 
                     + str(self._dim) + ' != ' + str(other.shape))
    
        C._A = self._A * other
        
        try:
            C._A = np.array(cas.evalf(C._A))
        except: 
            C._casadi = True
                    
        return C
    
    def __rmul__(self, other):
        if isinstance(other, (int, float)):
            return self.__mul__(other)
    
    def __matmul__(self, other): 
        
        if isinstance(other, array):
            if self._dim[1] != other._dim[0]:
                raise ValueError(
                    'inner dimensions are note the same - ' 
                    + str(self._dim[1]) + '!=' + str(other._dim[0]))
            dim = (self._dim[0], other._dim[1])
            other = other._A
            
        elif isinstance(other, np.ndarray):
            if len(other.shape) == 1:
                dim = (self._dim[0], 1)
                other = np.reshape(other, (other.shape[0], 1))
            else:
                dim = (self._dim[0], other.shape[1])
            if self._dim[1] != other.shape[0]:
                raise ValueError(
                    'inner dimensions are note the same - ' 
                    + str(self._dim[1]) + '!=' + str(other.shape[0]))
                
        elif isinstance(other, (cas.MX, cas.SX, cas.DM)):
            if self._dim[1] != other.size()[0]:
                raise ValueError(
                    'inner dimensions are note the same - ' 
                    + str(self._dim[1]) + '!=' + str(other.size()[0]))
            dim = (self._dim[0], other.size()[1])
            
        else: 
            raise TypeError(
                'input parameter must be of type ' + 
                'array, numpy.ndarray, ' 
                + 'casadi.MX, casadi.SX or casadi.DX - not ' 
                + str(type(other)))
        
        C = array(dim)
        
        C._A = self._A @ other
        
        try:
            C._A = np.array(cas.evalf(C._A))
        except: 
            C._casadi = True
        
        return C
    
    def __rmatmul__(self, other): 
    
        if isinstance(other, array):
            if self._dim[0] != other._dim[0]:
                raise ValueError(
                    'inner dimensions are note the same - ' 
                    + str(self._dim[0]) + 
                    '!=' + str(other._dim[1]))
            dim = (other._dim[0], self._dim[1])
            other = other._A
            
        elif isinstance(other, np.ndarray):
            if len(other.shape) == 1:
                dim = (1, self._dim[1])
                other = np.reshape(other, (1, other.shape[1]))
            else:
                dim = (other.shape[0], self._dim[1])
            if self._dim[0] != other.shape[1]:
                raise ValueError(
                    'inner dimensions are note the same - ' 
                    + str(self._dim[0]) + '!=' + str(other.shape[1]))
                
        elif isinstance(other, (cas.MX, cas.SX, cas.DM)):
            if self._dim[0] != other.size()[1]:
                raise ValueError(
                    'inner dimensions are note the same - ' 
                    + str(self._dim[0]) + '!=' + str(other.size()[1]))
            dim = (other.size()[0], self._dim[1])
            
        else: 
            raise TypeError(
                'input parameter must be of type' + 
                ' array, numpy.ndarray,' 
                + ' casadi.MX, casadi.SX or casadi.DX - not ' 
                + str(type(other)))
        
        C = array(dim)
        
        C._A = other @ self._A 
        
        try:
            C._A = np.array(cas.evalf(C._A))
        except: 
            C._casadi = True
        
        return C
    
    def __pow__(self, n): 
        
        if not isinstance(n, (int, float)):
            raise TypeError(
                'input parameter must be of type integer or float - not ' 
                + str(type(n)))
            
        return array(self._A**n)
    
    def __getitem__(self, key): 
    
        if isinstance(key, (int, slice)):
            key = (key,0)
        
        try:
            range(self._dim[0])[key[0]]
        except:
            raise IndexError('Index one is out of range: ' 
                             + str(key[0]) + ' not in range [0,' 
                             + str(self._dim[0]-1) + ']')
            
        try:
            range(self._dim[1])[key[1]]
        except:
            raise IndexError('Index one is out of range: ' 
                             + str(key[1]) + ' not in range [0,' 
                             + str(self._dim[1]-1) + ']')
        
        if isinstance(key,tuple):
            if len(key) != 2:
                raise ValueError(
                    'input parameter of type tuple has to be two dimensional')
            if (not isinstance(key[0], (int, slice)) or not 
                isinstance(key[1], (int, slice))):
                raise TypeError(
                    'input parameter must be of type integer or' 
                    + ' tuple of integers or slices')
            
            y = self._A[key[0],key[1]]
            
            if np.isscalar(y):
                return y
            elif isinstance(y, (cas.MX, cas.SX, cas.DM)) and y.is_scalar():
                try:
                    y = float(cas.evalf(y))
                except: None
                return y
            
            return array(y)
        
        else:
            raise TypeError(
                'input parameter must be of type integer or tuple - not ' 
                + str(type(key)))
        
    def __setitem__(self, key, value):
    
        if not isinstance(value, 
                          (array, np.ndarray, cas.MX, cas.SX, cas.DM, 
                           int, float, np.int64)):
             raise TypeError(
                 'input parameter must be of type' 
                 + ' array int, float, numpy.ndarray,' 
                 + ' casadi.MX, casadi.SX or casadi.DX - not ' 
                 + str(type(value)))
        
        if isinstance(key, int):
            key = (key,0)
            
        try:
            range(self._dim[0])[key[0]]
        except:
            raise IndexError('Index one is out of range: ' 
                             + str(key[0]) + ' not in range [0,' 
                             + str(self._dim[0]-1) + ']')
            
        try:
            range(self._dim[1])[key[1]]
        except:
            raise IndexError('Index one is out of range: ' 
                             + str(key[1]) + ' not in range [0,' 
                             + str(self._dim[1]-1) + ']')
        
        if isinstance(value, array):
            value = value._A
            
        if isinstance(key,tuple):
            
            if len(key) != 2:
                raise ValueError(
                    'key of type tuple has to be two dimensional')
            if (not isinstance(key[0], (int, slice)) 
                or not isinstance(key[1], (int, slice))):
                raise TypeError(
                    'key must be of type integer or' 
                    + ' tuple of integers or slices')
                
        else:
            raise TypeError(
                'key must be of type integer or tuple - not ' 
                + str(type(key)))
            
            
        if np.isscalar(value):
           value = float(value)
        elif (isinstance(value, (cas.MX, cas.SX, cas.DM)) and 
              value.is_scalar()):
            try:
                value = float(cas.evalf(value))
                
            except:
                if self._casadi == False:
                    self._A = cas.MX(self._A)
                    self._casadi = True
        else:
             if self._casadi == False:
                 self._A = cas.MX(self._A)
                 self._casadi = True
        
        try:
            self._A[key[0],key[1]] = value
        except: 
            self._A[key[0],key[1]] = value.flatten()
            
        try:
            self._A = np.array(cas.evalf(self._A))
            self._casadi = False
        except:
            self._casadi = True
        
    def _copy(self, other): 
        """Creats a new array object out of a casadi, numpy or array object.

        Parameters
        ----------
        other : array, casadi.MX, casadi.SX, casadi.DM, list or numpy.ndarray
            Object from which the array is created.

        """
        
        if isinstance(other, array):
            if isinstance(other._A, np.ndarray):
                self._A = other._A
            else:
                self._A = other._A
            self._dim = other._dim
            self._casadi = other._casadi
            
        elif isinstance(other, np.ndarray):
            if len(other.shape) == 1:
                self._dim = (other.shape[0], 1)
                self._A = np.reshape(other, self._dim)
            else:
                self._dim = other.shape
                self._A = other
            self._casadi = False
        
        elif isinstance(other, (cas.MX, cas.SX, cas.DM)):
            self._dim = other.size()
            self._A = other
            try:
                self._A = np.array(cas.evalf(self._A))
                self._casadi = False
            except:
                self._casadi = True
                    
        elif isinstance(other, list):
            if len(np.shape(other)) == 1:
                self._dim = (np.shape(other)[0], 1)
                other_ = []
                for i in range(self._dim[0]):
                    other_ += [[other[i]]]
                other = other_
            else:
                self._dim = np.shape(other)
            try:
                self._A = np.array(other)
                self._casadi = False
            except:
                self._A = cas.vertcat(*[cas.horzcat(*row) for row in other])
                self._casadi = True
                    
        else: 
            raise TypeError(
                'input parameter must be of type array, list, numpy.ndarray,' 
                + ' casadi.DM or casadi.SX, casadi.MX - not ' 
                + str(type(other)))
        
    def fill(self, a):
        """Fill all entries with one value.

        Parameters
        ----------
        a : int or float
            Value that all entries should take.

        """
        
        if isinstance(a, (float, int)):
            self._A = np.ones(self._dim)*a
            self._casadi = False
            
        else: 
            raise TypeError(
                'input paramter must be of type integer or float - not ' 
                + str(type(a)))
                    
    def flatten(self):
        """Flat array to one dimension.

        Returns
        -------
        y : array
            Flatten array with dimension (1,n).

        """
        
        y = array(self._A)
        dim2 = y._dim[0]*y._dim[1]
        dim1 = 1
        y._dim = (dim1,dim2)
        if isinstance(y._A, np.ndarray):
            y._A = np.reshape(y._A, y._dim)
        if isinstance(y._A, (cas.MX, cas.SX, cas.DM)):
            y._A = cas.reshape(y._A, y._dim)
            
        return y
    
    def transpose(self):
        """Transpose array.

        Returns
        -------
        y : array
            Transposed array.

        """
        
        y = array(self)
        dim2 = y._dim[0]
        dim1 = y._dim[1]
        y._dim = (dim1,dim2)
        if isinstance(y._A, np.ndarray):
            y._A = y._A.T
        elif isinstance(y._A, (cas.MX, cas.SX, cas.DM)):
            y._A = cas.transpose(y._A)
            
        return y
    
    
def reshape(a, new_size):
    """Reshape an array to a new size.

    Parameters
    ----------
    a : array
        .
    new_size : tuple
        New shape.

    Returns
    -------
    array
        An array instance with the new shape.

    """
    
    if not isinstance(new_size, tuple):
        raise TypeError(
            'new_size must be tuple - not ' 
            + str(type(new_size)))
    if not isinstance(new_size[0], int) or not isinstance(new_size[1], int):
        raise TypeError('new_size must be tuple of inegers')
    if not len(new_size) == 2:
        raise TypeError('new_size must be two dimensional')
        
    if isinstance(a, array):
        a = a._A
    else: 
        raise TypeError(
            'input paramter a must be of type array,- not ' + str(type(a)))
            
    if isinstance(a, np.ndarray):
        A = np.reshape(a,new_size)
    elif isinstance(a, (cas.MX, cas.SX, cas.DM)):
        A = cas.reshape(a, new_size)
    
    return array(A)

def convert(a, dtype='auto'):
    """Convert a numpy-, casadi- or nMPyC-array to another of these intances.

    Parameters
    ----------
    a : array, cas.MX, cas.SX, cas.DM or numpy.ndarray
        Array which should be converted.
    dtype : str, optional
        Name of the class to which the array will be converted. 
        The default is 'auto'.

    Returns
    -------
    numpy.ndarray, cas.MX, cas.SX, cas.DM
        The converted object.

    """
    
    if not isinstance(dtype, str):
        raise TypeError('dtype must be a string - not ' + str(type(dtype)))
    if dtype not in ['auto', 'numpy', 'casadi.DM', 'casadi.SX', 'casadi.MX']:
        raise ValueError('can not convert array to ' + dtype)
    
    if isinstance(a, array):
        a = a._A
        
    if isinstance(a, np.ndarray):
        if dtype == 'numpy':
            return a
        elif dtype == 'casadi.DM':
            return cas.DM(a)
        elif dtype == 'casadi.SX':
            return cas.SX(a)
        elif dtype == 'casadi.MX':
            return cas.MX(a)
        else: return a
    elif isinstance(a, cas.DM):
        if dtype == 'numpy':
            return np.array(a)
        elif dtype == 'casadi.DM':
            return a
        elif dtype == 'casadi.SX':
            return cas.SX(a)
        elif dtype == 'casadi.MX':
            return cas.MX(a)
        else: return a
    elif isinstance(a, cas.SX):
        if dtype == 'numpy':
            return np.array(cas.evalf(a))
        elif dtype == 'casadi.DM':
            return cas.evalf(a)
        elif dtype == 'casadi.SX':
            return a
        elif dtype == 'casadi.MX':
            return cas.MX(a)
        else: return a
    elif isinstance(a, cas.MX):
        if dtype == 'numpy':
            return np.array(cas.evalf(a))
        elif dtype == 'casadi.DM':
            return cas.evalf(a)
        elif dtype == 'casadi.SX':
            return cas.SX(cas.evalf(a))
        elif dtype == 'casadi.MX':
            return a
        else: return a
        
    else:
        raise TypeError(
            'input parameter a must be of type array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(a)))
    
    return a

def concatenate(arrays, axis = 0):
    """Join a sequence of arrays along an existing axis.

    Parameters
    ----------
    arrays : tuple of casadi.MX, casadi.SX, casadi.DM or numpy.ndarrays
        Sequence of arrays which will be concatenated. 
        The arrays must have the same shape, except in the dimension 
        corresponding to axis.
    axis : int, optional
        The axis along which the arrays will be joined. The default is 0.

    Returns
    -------
    array
        The concatenated array.

    """
    
    casadi = False
    
    if not isinstance(arrays, tuple):
        raise TypeError(
            'arrays hmust be of type tuple - not ' + str(type(arrays)))
        
    if not isinstance(axis, int):
        raise TypeError('axis must be an integer - not ' + str(type(axis)))
    
    array_list = []
    for i in range(len(arrays)):
        if isinstance(arrays[i], array):
            array_list += [arrays[i].A]
        else:
            array_list += [arrays[i]]
        if isinstance(array_list[i], (cas.MX, cas.SX, cas.DM)):
            casadi = True
        elif not isinstance(array_list[i], np.ndarray):
            raise ValueError(
                'The elements of the arrays sequence must be of type array,' + 
                ' casadi.MX, casadi.SX, casadi.DM or numpy.ndarray - not ' 
                + str(type(arrays[i])))
            
    arrays = tuple(array_list)
        
    if axis == 0:
        if casadi:
            return array(cas.vertcat(*arrays))
        else:
            return array(np.concatenate(arrays, axis=0))
    elif axis == 1:
        if casadi:
            return array(cas.horzcat(*arrays))
        else:
            return array(np.concatenate(arrays, axis=1))
    else:
        raise ValueError(
            'axis must be 0 (first axis) or 1 (second axis) - not ' 
            + str(axis))
    
def eye(dim):
    """Creates an array defining the idendity.

    Parameters
    ----------
    dim : int
        Dimnension of the idendity matrix.

    Raises
    ------
    ValueError
        If the given dimension is not supported.
    TypeError
        If an input parameter has not the right type.

    Returns
    -------
    y : array
        Idendity matrix as an instance of array.

    """
    
    if isinstance(dim, int):
        if dim <= 0:
            raise ValueError('dimension must be graeter than zero')
    else: 
        raise TypeError(
            'dimension must be an integer - not ' 
            + str(type(dim)))
    
    y = array((dim, dim))
    y._A = np.eye(dim)
    return y

def zeros(dim):
    """Creates an array with only zero entries.

    Parameters
    ----------
    dim : int or tuple
        Dimension of the array.

    Raises
    ------
    ValueError
        If the given dimension is not supported.
    TypeError
        If the given dimension has not the right type.

    Returns
    -------
    y : array
        An array of the given dimension with only zero entries.

    """
    
    if isinstance(dim, int): 
        None
    elif isinstance(dim, tuple):
        if len(dim) != 2:
            raise ValueError('dim must be two dimensional')
        if not isinstance(dim[0], int) or not isinstance(dim[0], int):
            raise TypeError('dimension must be tuple of integers')
        if dim[0] <= 0 or dim[1] <= 0:
            raise ValueError('dimension must be graeter than zero')
    else: 
        raise TypeError(
            'dimension must be an integer or tuple - not ' 
            + str(type(dim)))
    
    y = array(dim)
    return y

def ones(dim):
    """Creates an array with only entries equal to one.

    Parameters
    ----------
    dim : int or tuple
        Dimension of the array.

    Raises
    ------
    ValueError
        If the given dimension is not supported.
    TypeError
        If the given dimension has not the right type.

    Returns
    -------
    y : array
        An array of the given dimension with only entries equal to one.

    """

    if isinstance(dim, int): None
    elif isinstance(dim, tuple):
        if len(dim) != 2:
            raise ValueError('dim must be two dimensional')
        if not isinstance(dim[0], int) or not isinstance(dim[0], int):
            raise TypeError('dimension must be tuple of integers')
        if dim[0] <= 0 or dim[1] <= 0:
            raise ValueError('dimension must be graeter than zero')
    else: 
        raise TypeError(
            'dimension must be an integer or tuple - not ' 
            + str(type(dim)))
    
    y = array(dim)
    y.fill(1)
    return y

def diag(x):
    """Creates an diagonal matrix from a given vector.

    Parameters
    ----------
    x : array, numpy.ndarray, cas.MX, cas.SX or cas.DX, list
        Vector containing the diagonal entries of the matrix.

    Returns
    -------
    array
        Diagonal matrix with the desired diagonal elements.

    """
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (np.ndarray)):
        y = np.diag(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM, list)):
        y = cas.diag(x_)
        
    else: 
        raise TypeError(
            'input must be of type array, numpy.ndarray, list,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
        
    return array(y)

def log(x):
    """Calculates the natural logarithm of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int,float,np.ndarray)):
        y = np.log(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.log(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def exp(x):
    """Calculates the exponential of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.exp(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.exp(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def sin(x):
    """Calculates the sinus of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.sin(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.sin(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def sinh(x):
    """Calculates the sinus hyperbolicus of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.sinh(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.sinh(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def arcsin(x):
    """Calculates the arcussinus of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.arcsin(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.arcsin(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def arcsinh(x):
    """Calculates the arcussinus hyperbolicus of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.arcsinh(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.arcsinh(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def cos(x):
    """Calculates the cosinus of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.cos(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.cos(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def cosh(x):
    """Calculates the cosinus hyperbolicus of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.cosh(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.cosh(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def arccos(x):
    """Calculates the arcuscosinus of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.arccos(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.arccos(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def arccosh(x):
    """Calculates the arcuscosinus hypernolicus of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.arccosh(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.arccosh(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def tan(x):
    """Calculates the tangens of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.tan(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.tan(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def tanh(x):
    """Calculates the tangens hyperblicus of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.tan(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.tan(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def arctan(x):
    """Calculates the arcustangens of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.arctan(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.arctan(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def arctanh(x):
    """Calculates the arcustangens hyperbolicus of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.arctanh(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.arctanh(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def sqrt(x):
    """Calculates the square root of a given number or array"""
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.sqrt(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.sqrt(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def power(x,n):
    """Calculates the (elementwise) n-th power of a number or array.

    Parameters
    ----------
    x : int, float, numpy.ndarray, cas.MS, cas.SX or cas.DM
        Number or array of which the n-th power should be computed.
    n : int or float
        Number defining the exponent.

    """
    
    if not isinstance(n, (int, float)):
        raise TypeError(
            'input n must be of type integer or float - not ' 
            + str(type(n)))
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.power(x_,n)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.power(x_,n)
        
    else: 
        raise TypeError(
            'input x must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def matrix_power(x,n):
    """Raises a square matrix to the n-th power.

    Parameters
    ----------
    x : int, float, numpy.ndarray, cas.MS, cas.SX or cas.DM
        Number or array of which the n-th power should be computed.
    n : int or float
        Number defining the exponent.

    """
    
    if not isinstance(n, (int, float)):
        raise TypeError(
            'input n must be of type integer or float - not ' 
            + str(type(n)))
    
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (np.ndarray)):
        y = np.linalg.matrix_power(x_,n)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.mpower(x_,n)
        
    else: 
        raise TypeError(
            'input x must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y

def abs(x):
    """Calculates the absolute value of a number or array."""

    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (int, float, np.ndarray)):
        y = np.absolute(x_)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        y = cas.fabs(x_)
        
    else: 
        raise TypeError(
            'input must be of type integer, float, array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' + str(type(x)))
    
    if isinstance(x, array):
        y = array(y)

    return y
    
def norm(x,order=None):
    """Returns the norm of a vector or matrix.

    Parameters
    ----------
    x : array, numpy.ndarray, cas.MX, cas.SX or cas.DM
        Vector or matrix of which the norm should be calculated.
    order : number or str, optional
        String defining the type of the norm. 
        Posiible values are 1, 2, 'fro' or 'inf'. 
        The default is None.

    """
    if order is not None and order not in [1,2,'fro',inf]:
        raise ValueError(
            'order must be one 1,2, fro or inf - not ' 
            + str(order))
        
    if isinstance(x, array):
        x_ = x._A
    else:
        x_ = x
        
    if isinstance(x_, (np.ndarray)):
        y = LA.norm(x_,order)
        
    elif isinstance(x_, (cas.MX, cas.SX, cas.DM)):
        if order == 1:
            y = cas.norm_1(x_)
        elif order == 2:
            y = cas.norm_2(x_)
        elif order == 'fro':
            y = cas.norm_fro(x_)
        elif order == inf:
            y = cas.norm_inf(x_)
        
    else: 
        raise TypeError(
            'input must be of type array, numpy.ndarray,' 
            + ' casadi.DM, casadi.SX or casadi.MX - not ' 
            + str(type(x)))

    return y

def max(*args):
    """Return the maximal value of the arguments"""
       
    for i in range(len(args)):
        if isinstance(args[i], array):
            args[i] = args[i]._A
        if not isinstance(args[i], (int, float, 
                                    np.ndarray, cas.MX, cas.SX, cas.DM)):
            raise TypeError(
                'input must be of type integer, float, array, numpy.ndarray,' 
                + ' casadi.DM, casadi.SX or casadi.MX - not ' 
                + str(type(args[i])))

    return cas.fmax(*args)

def min(*args):
    """Returns the minimal value of the arguments"""
    
    for i in range(len(args)):
        if isinstance(args[i], array):
            args[i] = args[i]._A
        if not isinstance(args[i], (int, float, 
                                    np.ndarray, cas.MX, cas.SX, cas.DM)):
            raise TypeError(
                'input must be of type integer, float, array, numpy.ndarray,' 
                + ' casadi.DM, casadi.SX or casadi.MX - not ' 
                + str(type(args[i])))

    return cas.fmin(*args)
    
        
if __name__ == '__main__':        
    # a = np.array((1,2,3))
    # a = cas.DM((3,3))
    # print(a[:,0])
    # a = array(a)
    # a[:,0] = np.array([[2,2]])
    # print(a)
    # print(a[:,0])
    
    # b = concatenate((a,a))
    # print(b)
    
    # b = concatenate((a,a),axis=1)
    # print(b)
    
    # print(array.__doc__)
    
    A = zeros(10)
    print(A[10])
